"""
Injected database interface dependencies
"""

import asyncio
from contextlib import ExitStack, contextmanager
from functools import wraps
from typing import Callable, Type, TypeVar

from typing_extensions import Concatenate, ParamSpec

from prefect.server.database.configurations import (
    AioSqliteConfiguration,
    AsyncPostgresConfiguration,
    BaseDatabaseConfiguration,
)
from prefect.server.database.interface import PrefectDBInterface
from prefect.server.database.orm_models import (
    AioSqliteORMConfiguration,
    AsyncPostgresORMConfiguration,
    BaseORMConfiguration,
)
from prefect.server.database.query_components import (
    AioSqliteQueryComponents,
    AsyncPostgresQueryComponents,
    BaseQueryComponents,
)
from prefect.server.utilities.database import get_dialect
from prefect.settings import PREFECT_API_DATABASE_CONNECTION_URL

MODELS_DEPENDENCIES = {
    "database_config": None,
    "query_components": None,
    "orm": None,
    "interface_class": None,
}

P = ParamSpec("P")
R = TypeVar("R")


def provide_database_interface() -> PrefectDBInterface:
    """
    Get the current Prefect REST API database interface.

    If components of the interface are not set, defaults will be inferred
    based on the dialect of the connection URL.
    """
    connection_url = PREFECT_API_DATABASE_CONNECTION_URL.value()

    database_config = MODELS_DEPENDENCIES.get("database_config")
    query_components = MODELS_DEPENDENCIES.get("query_components")
    orm = MODELS_DEPENDENCIES.get("orm")
    interface_class = MODELS_DEPENDENCIES.get("interface_class")
    dialect = get_dialect(connection_url)

    if database_config is None:
        if dialect.name == "postgresql":
            database_config = AsyncPostgresConfiguration(connection_url=connection_url)
        elif dialect.name == "sqlite":
            database_config = AioSqliteConfiguration(connection_url=connection_url)
        else:
            raise ValueError(
                "Unable to infer database configuration from provided dialect. Got"
                f" dialect name {dialect.name!r}"
            )

        MODELS_DEPENDENCIES["database_config"] = database_config

    if query_components is None:
        if dialect.name == "postgresql":
            query_components = AsyncPostgresQueryComponents()
        elif dialect.name == "sqlite":
            query_components = AioSqliteQueryComponents()
        else:
            raise ValueError(
                "Unable to infer query components from provided dialect. Got dialect"
                f" name {dialect.name!r}"
            )

        MODELS_DEPENDENCIES["query_components"] = query_components

    if orm is None:
        if dialect.name == "postgresql":
            orm = AsyncPostgresORMConfiguration()
        elif dialect.name == "sqlite":
            orm = AioSqliteORMConfiguration()
        else:
            raise ValueError(
                "Unable to infer orm configuration from provided dialect. Got dialect"
                f" name {dialect.name!r}"
            )

        MODELS_DEPENDENCIES["orm"] = orm

    if interface_class is None:
        interface_class = PrefectDBInterface

    return interface_class(
        database_config=database_config,
        query_components=query_components,
        orm=orm,
    )


def inject_db(fn: Callable) -> Callable:
    """
    Decorator that provides a database interface to a function.

    The decorated function _must_ take a `db` kwarg and if a db is passed
    when called it will be used instead of creating a new one.

    If the function is a coroutine function, the wrapper will await the
    function's result. Otherwise, the wrapper will call the function
    normally.
    """

    def inject(kwargs):
        if "db" not in kwargs or kwargs["db"] is None:
            kwargs["db"] = provide_database_interface()

    @wraps(fn)
    async def async_wrapper(*args, **kwargs):
        inject(kwargs)
        return await fn(*args, **kwargs)

    @wraps(fn)
    def sync_wrapper(*args, **kwargs):
        inject(kwargs)
        return fn(*args, **kwargs)

    if asyncio.iscoroutinefunction(fn):
        return async_wrapper

    return sync_wrapper


def db_injector(
    func: Callable[Concatenate[PrefectDBInterface, P], R],
) -> Callable[P, R]:
    """
    Decorator to inject a PrefectDBInterface instance as the first positional
    argument to the decorated function.

    Unlike `inject_db`, which injects the database connection as a keyword
    argument, `db_injector` adds it explicitly as the first positional
    argument. This change enhances type hinting by making the dependency on
    PrefectDBInterface explicit in the function signature.

    Args:
        func: The function to decorate, which can be either synchronous or
        asynchronous.

    Returns:
        A wrapped function with the PrefectDBInterface instance injected as the
        first argument, preserving the original function's other parameters and
        return type.
    """

    @wraps(func)
    def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        db = provide_database_interface()
        return func(db, *args, **kwargs)

    @wraps(func)
    async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        db = provide_database_interface()
        return await func(db, *args, **kwargs)  # type: ignore

    if asyncio.iscoroutinefunction(func):
        return async_wrapper  # type: ignore
    else:
        return sync_wrapper


@contextmanager
def temporary_database_config(tmp_database_config: BaseDatabaseConfiguration):
    """
    Temporarily override the Prefect REST API database configuration.
    When the context is closed, the existing database configuration will
    be restored.

    Args:
        tmp_database_config: Prefect REST API database configuration to inject.

    """
    starting_config = MODELS_DEPENDENCIES["database_config"]
    try:
        MODELS_DEPENDENCIES["database_config"] = tmp_database_config
        yield
    finally:
        MODELS_DEPENDENCIES["database_config"] = starting_config


@contextmanager
def temporary_query_components(tmp_queries: BaseQueryComponents):
    """
    Temporarily override the Prefect REST API database query components.
    When the context is closed, the existing query components will
    be restored.

    Args:
        tmp_queries: Prefect REST API query components to inject.

    """
    starting_queries = MODELS_DEPENDENCIES["query_components"]
    try:
        MODELS_DEPENDENCIES["query_components"] = tmp_queries
        yield
    finally:
        MODELS_DEPENDENCIES["query_components"] = starting_queries


@contextmanager
def temporary_orm_config(tmp_orm_config: BaseORMConfiguration):
    """
    Temporarily override the Prefect REST API ORM configuration.
    When the context is closed, the existing orm configuration will
    be restored.

    Args:
        tmp_orm_config: Prefect REST API ORM configuration to inject.

    """
    starting_orm_config = MODELS_DEPENDENCIES["orm"]
    try:
        MODELS_DEPENDENCIES["orm"] = tmp_orm_config
        yield
    finally:
        MODELS_DEPENDENCIES["orm"] = starting_orm_config


@contextmanager
def temporary_interface_class(tmp_interface_class: Type[PrefectDBInterface]):
    """
    Temporarily override the Prefect REST API interface class When the context is closed,
    the existing interface will be restored.

    Args:
        tmp_interface_class: Prefect REST API interface class to inject.

    """
    starting_interface_class = MODELS_DEPENDENCIES["interface_class"]
    try:
        MODELS_DEPENDENCIES["interface_class"] = tmp_interface_class
        yield
    finally:
        MODELS_DEPENDENCIES["interface_class"] = starting_interface_class


@contextmanager
def temporary_database_interface(
    tmp_database_config: BaseDatabaseConfiguration = None,
    tmp_queries: BaseQueryComponents = None,
    tmp_orm_config: BaseORMConfiguration = None,
    tmp_interface_class: Type[PrefectDBInterface] = None,
):
    """
    Temporarily override the Prefect REST API database interface.

    Any interface components that are not explicitly provided will be
    cleared and inferred from the Prefect REST API database connection string
    dialect.

    When the context is closed, the existing database interface will
    be restored.

    Args:
        tmp_database_config: An optional Prefect REST API database configuration to inject.
        tmp_orm_config: An optional Prefect REST API ORM configuration to inject.
        tmp_queries: Optional Prefect REST API query components to inject.
        tmp_interface_class: Optional database interface class to inject

    """
    with ExitStack() as stack:
        stack.enter_context(
            temporary_database_config(tmp_database_config=tmp_database_config)
        )
        stack.enter_context(temporary_query_components(tmp_queries=tmp_queries))
        stack.enter_context(temporary_orm_config(tmp_orm_config=tmp_orm_config))
        stack.enter_context(
            temporary_interface_class(tmp_interface_class=tmp_interface_class)
        )
        yield


def set_database_config(database_config: BaseDatabaseConfiguration):
    """Set Prefect REST API database configuration."""
    MODELS_DEPENDENCIES["database_config"] = database_config


def set_query_components(query_components: BaseQueryComponents):
    """Set Prefect REST API query components."""
    MODELS_DEPENDENCIES["query_components"] = query_components


def set_orm_config(orm_config: BaseORMConfiguration):
    """Set Prefect REST API orm configuration."""
    MODELS_DEPENDENCIES["orm"] = orm_config


def set_interface_class(interface_class: Type[PrefectDBInterface]):
    """Set Prefect REST API interface class."""
    MODELS_DEPENDENCIES["interface_class"] = interface_class



import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional

from opentelemetry.propagators.textmap import Setter
from opentelemetry.trace import (
    Status,
    StatusCode,
    get_tracer,
)

import prefect
from prefect.client.schemas import TaskRun
from prefect.client.schemas.objects import State
from prefect.types import KeyValueLabels

if TYPE_CHECKING:
    from opentelemetry.trace import Tracer


class OTELSetter(Setter[KeyValueLabels]):
    """
    A setter for OpenTelemetry that supports Prefect's custom labels.
    """

    def set(self, carrier: KeyValueLabels, key: str, value: str) -> None:
        carrier[key] = value


@dataclass
class RunTelemetry:
    """
    A class for managing the telemetry of runs.
    """

    _tracer: "Tracer" = field(
        default_factory=lambda: get_tracer("prefect", prefect.__version__)
    )
    _span = None

    def start_span(
        self,
        task_run: TaskRun,
        parameters: Optional[Dict[str, Any]] = None,
        labels: Optional[Dict[str, Any]] = None,
    ):
        """
        Start a span for a task run.
        """
        if parameters is None:
            parameters = {}
        if labels is None:
            labels = {}
        parameter_attributes = {
            f"prefect.run.parameter.{k}": type(v).__name__
            for k, v in parameters.items()
        }
        self._span = self._tracer.start_span(
            name=task_run.name,
            attributes={
                "prefect.run.type": "task",
                "prefect.run.id": str(task_run.id),
                "prefect.tags": task_run.tags,
                **parameter_attributes,
                **labels,
            },
        )

    def end_span_on_success(self, terminal_message: str) -> None:
        """
        End a span for a task run on success.
        """
        if self._span:
            self._span.set_status(Status(StatusCode.OK), terminal_message)
            self._span.end(time.time_ns())
            self._span = None

    def end_span_on_failure(self, terminal_message: str) -> None:
        """
        End a span for a task run on failure.
        """
        if self._span:
            self._span.set_status(Status(StatusCode.ERROR, terminal_message))
            self._span.end(time.time_ns())
            self._span = None

    def record_exception(self, exc: Exception) -> None:
        """
        Record an exception on a span.
        """
        if self._span:
            self._span.record_exception(exc)

    def update_state(self, new_state: State) -> None:
        """
        Update a span with the state of a task run.
        """
        if self._span:
            self._span.add_event(
                new_state.name or new_state.type,
                {
                    "prefect.state.message": new_state.message or "",
                    "prefect.state.type": new_state.type,
                    "prefect.state.name": new_state.name or new_state.type,
                    "prefect.state.id": str(new_state.id),
                },
            )



"""
Access attributes of the current task run dynamically.

Note that if a task run cannot be discovered, all attributes will return empty values.

You can mock the runtime attributes for testing purposes by setting environment variables
prefixed with `PREFECT__RUNTIME__TASK_RUN`.

Available attributes:
    - `id`: the task run's unique ID
    - `name`: the name of the task run
    - `tags`: the task run's set of tags
    - `parameters`: the parameters the task was called with
    - `run_count`: the number of times this task run has been run
    - `task_name`: the name of the task
"""

import os
from typing import Any, Dict, List, Optional

from prefect.context import TaskRunContext

__all__ = ["id", "tags", "name", "parameters", "run_count", "task_name"]


type_cast = {
    bool: lambda x: x.lower() == "true",
    int: int,
    float: float,
    str: str,
    # for optional defined attributes, when real value is NoneType, use str
    type(None): str,
}


def __getattr__(name: str) -> Any:
    """
    Attribute accessor for this submodule; note that imports also work with this:

        from prefect.runtime.task_run import id
    """

    func = FIELDS.get(name)

    # if `name` is an attribute but it is mocked through environment variable, the mocked type will be str,
    # which might be different from original one. For consistency, cast env var to the same type
    env_key = f"PREFECT__RUNTIME__TASK_RUN__{name.upper()}"

    if func is None:
        if env_key in os.environ:
            return os.environ[env_key]
        else:
            raise AttributeError(f"{__name__} has no attribute {name!r}")

    real_value = func()
    if env_key in os.environ:
        mocked_value = os.environ[env_key]
        # cast `mocked_value` to the same type as `real_value`
        try:
            cast_func = type_cast[type(real_value)]
            return cast_func(mocked_value)
        except KeyError:
            raise ValueError(
                "This runtime context attribute cannot be mocked using an"
                " environment variable. Please use monkeypatch instead."
            )
    else:
        return real_value


def __dir__() -> List[str]:
    return sorted(__all__)


def get_id() -> str:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is not None:
        return str(task_run_ctx.task_run.id)


def get_tags() -> List[str]:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is None:
        return []
    else:
        return task_run_ctx.task_run.tags


def get_run_count() -> int:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is None:
        return 0
    else:
        return task_run_ctx.task_run.run_count


def get_name() -> Optional[str]:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is None:
        return None
    else:
        return task_run_ctx.task_run.name


def get_task_name() -> Optional[str]:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is None:
        return None
    else:
        return task_run_ctx.task.name


def get_parameters() -> Dict[str, Any]:
    task_run_ctx = TaskRunContext.get()
    if task_run_ctx is not None:
        return task_run_ctx.parameters
    else:
        return {}


FIELDS = {
    "id": get_id,
    "tags": get_tags,
    "name": get_name,
    "parameters": get_parameters,
    "run_count": get_run_count,
    "task_name": get_task_name,
}



"""Migrates block schemas with new secrets fields

Revision ID: 4cdc2ba709a4
Revises: 813ddf14e2de
Create Date: 2022-07-06 15:25:28.026455

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4cdc2ba709a4"
down_revision = "813ddf14e2de"
branch_labels = None
depends_on = None


BLOCKS_TO_MIGRATE = [
    {
        "BLOCK_TYPE_NAME": "S3 Storage",
        "OLD_CHECKSUM": (
            "sha256:3ffda32926202a34fc30d7618e4792f71088d37fbcf92c18d070104ca6da7431"
        ),
        "NEW_CHECKSUM": (
            "sha256:68ed6efe6ab724c5f36519803012af68a22965eccfa5944c94fa809b8b9a6e04"
        ),
        "CAPABILITIES": ["readable", "storage", "writeable"],
        "NEW_FIELDS": {
            "title": "S3StorageBlock",
            "description": "Store data in an AWS S3 bucket.",
            "type": "object",
            "properties": {
                "bucket": {"title": "Bucket", "type": "string"},
                "aws_access_key_id": {"title": "Aws Access Key Id", "type": "string"},
                "aws_secret_access_key": {
                    "title": "Aws Secret Access Key",
                    "type": "string",
                    "writeOnly": True,
                    "format": "password",
                },
                "aws_session_token": {"title": "Aws Session Token", "type": "string"},
                "profile_name": {"title": "Profile Name", "type": "string"},
                "region_name": {"title": "Region Name", "type": "string"},
            },
            "required": ["bucket"],
            "block_type_name": "S3 Storage",
            "secret_fields": ["aws_secret_access_key"],
            "block_schema_references": {},
        },
    },
    {
        "BLOCK_TYPE_NAME": "Azure Blob Storage",
        "OLD_CHECKSUM": (
            "sha256:4488e8f7d196f7627e3ead24ca136860f0a54d54f6c98533cf3ef2f4ba9cf51b"
        ),
        "NEW_CHECKSUM": (
            "sha256:2aef5e384a1f4a2d8dd0ff8c3b96d2c5eb5852462078b6915f7d756847341a42"
        ),
        "CAPABILITIES": ["readable", "storage", "writeable"],
        "NEW_FIELDS": {
            "title": "AzureBlobStorageBlock",
            "description": "Store data in an Azure blob storage container.",
            "type": "object",
            "properties": {
                "container": {"title": "Container", "type": "string"},
                "connection_string": {
                    "title": "Connection String",
                    "type": "string",
                    "writeOnly": True,
                    "format": "password",
                },
            },
            "required": ["container", "connection_string"],
            "block_type_name": "Azure Blob Storage",
            "secret_fields": ["connection_string"],
            "block_schema_references": {},
        },
    },
]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    meta_data = sa.MetaData()
    meta_data.reflect(connection)
    BLOCK_TYPE = meta_data.tables["block_type"]
    BLOCK_SCHEMA = meta_data.tables["block_schema"]
    BLOCK_DOCUMENT = meta_data.tables["block_document"]

    for block_migration_config in BLOCKS_TO_MIGRATE:
        block_type_result = connection.execute(
            sa.select(BLOCK_TYPE.c.id).where(
                BLOCK_TYPE.c.name == block_migration_config["BLOCK_TYPE_NAME"]
            )
        ).first()
        old_block_schema_result = connection.execute(
            sa.select(BLOCK_SCHEMA.c.id).where(
                BLOCK_SCHEMA.c.checksum == block_migration_config["OLD_CHECKSUM"]
            )
        ).first()
        # Only run migration for this block if the type and old schema already exist
        if block_type_result is not None and old_block_schema_result is not None:
            # Check if new version of the schema is present
            new_block_schema_result = connection.execute(
                sa.select(BLOCK_SCHEMA.c.id).where(
                    BLOCK_SCHEMA.c.checksum == block_migration_config["NEW_CHECKSUM"]
                )
            ).first()
            if new_block_schema_result is None:
                # Create new schema if not present
                connection.execute(
                    sa.insert(BLOCK_SCHEMA).values(
                        checksum=block_migration_config["NEW_CHECKSUM"],
                        fields=block_migration_config["NEW_FIELDS"],
                        block_type_id=block_type_result[0],
                        capabilities=block_migration_config["CAPABILITIES"],
                    )
                )
                new_block_schema_result = connection.execute(
                    sa.select(BLOCK_SCHEMA.c.id).where(
                        BLOCK_SCHEMA.c.checksum
                        == block_migration_config["NEW_CHECKSUM"]
                    )
                ).first()
            # Get all block documents that use the old block schema
            existing_block_documents_result = connection.execute(
                sa.select(BLOCK_DOCUMENT.c.id).where(
                    BLOCK_DOCUMENT.c.block_schema_id == old_block_schema_result[0]
                )
            ).all()
            # Update all block documents using the old block schema to use new block schema
            for block_document in existing_block_documents_result:
                connection.execute(
                    sa.update(BLOCK_DOCUMENT)
                    .where(BLOCK_DOCUMENT.c.id == block_document[0])
                    .values(block_schema_id=new_block_schema_result[0])
                )
            # Remove the old unused block schema
            connection.execute(
                sa.delete(BLOCK_SCHEMA).where(
                    BLOCK_SCHEMA.c.id == old_block_schema_result[0]
                )
            )
    # ### end Alembic commands ###


def downgrade():
    # Purely a data migration for 2.0b8. No downgrade necessary.
    pass



"""Work queue data migration

Revision ID: f98ae6d8e2cc
Revises: 0a1250a5aa25
Create Date: 2023-01-31 11:05:43.356002

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f98ae6d8e2cc"
down_revision = "0a1250a5aa25"
branch_labels = None
depends_on = None


def upgrade():
    # Create temporary indexes for migration
    op.execute(
        "CREATE INDEX IF NOT EXISTS ix_flow_run__work_queue_id_work_queue_name ON"
        " flow_run (work_queue_id, work_queue_name)"
    )
    op.execute(
        "CREATE INDEX IF NOT EXISTS ix_deployment__work_queue_id_work_queue_name ON"
        " deployment (work_queue_id, work_queue_name)"
    )

    # Create default agent work pool and associate all existing queues with it
    connection = op.get_bind()

    n_work_queues_without_a_work_pool = connection.execute(
        sa.text("SELECT COUNT(*) FROM work_queue WHERE work_pool_id IS NULL")
    ).fetchone()[0]

    if n_work_queues_without_a_work_pool > 0:
        connection.execute(
            sa.text(
                "INSERT INTO work_pool (name, type) VALUES ('default-agent-pool',"
                " 'prefect-agent')"
            )
        )

        default_pool_id = connection.execute(
            sa.text("SELECT id FROM work_pool WHERE name = 'default-agent-pool'")
        ).fetchone()[0]

        default_queue = connection.execute(
            sa.text("SELECT id FROM work_queue WHERE name = 'default'")
        ).fetchone()

        if not default_queue:
            connection.execute(
                sa.text(
                    "INSERT INTO work_queue (name, work_pool_id) VALUES ('default',"
                    " :default_pool_id)"
                ).params({"default_pool_id": default_pool_id}),
            )

        connection.execute(
            sa.text(
                "UPDATE work_queue SET work_pool_id = :default_pool_id WHERE work_pool_id"
                " IS NULL"
            ).params({"default_pool_id": default_pool_id}),
        )

        default_queue_id = connection.execute(
            sa.text(
                "SELECT id FROM work_queue WHERE name = 'default' and work_pool_id ="
                " :default_pool_id"
            ).params({"default_pool_id": default_pool_id}),
        ).fetchone()[0]

        connection.execute(
            sa.text(
                "UPDATE work_pool SET default_queue_id = :default_queue_id WHERE id ="
                " :default_pool_id"
            ).params(
                {
                    "default_pool_id": default_pool_id,
                    "default_queue_id": default_queue_id,
                }
            ),
        )

        # Set priority on all queues and update flow runs and deployments
        queue_rows = connection.execute(
            sa.text(
                "SELECT id, name FROM work_queue WHERE work_pool_id = :default_pool_id"
            ).params({"default_pool_id": default_pool_id}),
        ).fetchall()

        with op.get_context().autocommit_block():
            for enumeration, row in enumerate(queue_rows):
                connection.execute(
                    sa.text(
                        "UPDATE work_queue SET priority = :priority WHERE id = :id"
                    ).params({"priority": enumeration + 1, "id": row[0]}),
                )

                batch_size = 250

                while True:
                    result = connection.execute(
                        sa.text(
                            """
                            UPDATE flow_run 
                            SET work_queue_id=:id 
                            WHERE flow_run.id in (
                                SELECT id 
                                FROM flow_run 
                                WHERE flow_run.work_queue_id IS NULL and flow_run.work_queue_name=:name 
                                LIMIT :batch_size
                            )
                            """
                        ).params(
                            {"id": row[0], "name": row[1], "batch_size": batch_size}
                        ),
                    )
                    if result.rowcount <= batch_size:
                        break

                while True:
                    result = connection.execute(
                        sa.text(
                            """
                            UPDATE deployment 
                            SET work_queue_id=:id 
                            WHERE deployment.id in (
                                SELECT id 
                                FROM deployment 
                                WHERE deployment.work_queue_id IS NULL and deployment.work_queue_name=:name 
                                LIMIT :batch_size
                            )
                            """
                        ).params(
                            {"id": row[0], "name": row[1], "batch_size": batch_size}
                        ),
                    )
                    if result.rowcount <= batch_size:
                        break

    with op.batch_alter_table("work_queue", schema=None) as batch_op:
        batch_op.drop_constraint("uq_work_queue__name")
        batch_op.create_unique_constraint(
            op.f("uq_work_queue__work_pool_id_name"), ["work_pool_id", "name"]
        )
        batch_op.alter_column("work_pool_id", nullable=False)

    op.execute("DROP INDEX IF EXISTS ix_flow_run__work_queue_id_work_queue_name")
    op.execute("DROP INDEX IF EXISTS ix_deployment__work_queue_id_work_queue_name")


def downgrade():
    # Create temporary indexes for migration
    op.execute(
        "CREATE INDEX IF NOT EXISTS ix_flow_run__work_queue_id_work_queue_name ON"
        " flow_run (work_queue_id, work_queue_name)"
    )
    op.execute(
        "CREATE INDEX IF NOT EXISTS ix_deployment__work_queue_id_work_queue_name ON"
        " deployment (work_queue_id, work_queue_name)"
    )

    with op.batch_alter_table("work_queue", schema=None) as batch_op:
        batch_op.alter_column("work_pool_id", nullable=True)

    connection = op.get_bind()

    # Delete all non-default queues and pools
    default_pool_id_result = connection.execute(
        sa.text("SELECT id FROM work_pool WHERE name = 'default-agent-pool'")
    ).fetchone()
    if default_pool_id_result:
        default_pool_id = default_pool_id_result[0]
        connection.execute(
            sa.text(
                "DELETE FROM work_queue WHERE work_pool_id != :default_pool_id"
            ).params({"default_pool_id": default_pool_id})
        )
    queue_rows = connection.execute(
        sa.text("SELECT id, name FROM work_queue"),
    ).fetchall()

    with op.get_context().autocommit_block():
        for row in queue_rows:
            batch_size = 250

            while True:
                result = connection.execute(
                    sa.text(
                        """
                        UPDATE flow_run 
                        SET work_queue_id=NULL 
                        WHERE flow_run.id in (
                            SELECT id 
                            FROM flow_run 
                            WHERE flow_run.work_queue_id IS NOT NULL and flow_run.work_queue_id=:id 
                            LIMIT :batch_size
                        )
                        """
                    ).params({"id": row[0], "batch_size": batch_size}),
                )
                if result.rowcount <= batch_size:
                    break

            while True:
                result = connection.execute(
                    sa.text(
                        """
                        UPDATE deployment 
                        SET work_queue_id=NULL 
                        WHERE deployment.id in (
                            SELECT id 
                            FROM deployment 
                            WHERE deployment.work_queue_id IS NOT NULL and deployment.work_queue_id=:id
                            LIMIT :batch_size
                        )
                        """
                    ).params({"id": row[0], "batch_size": batch_size}),
                )
                if result.rowcount <= batch_size:
                    break

    connection.execute(sa.text("UPDATE work_queue SET work_pool_id = NULL"))

    connection.execute(sa.text("DELETE FROM work_pool"))

    with op.batch_alter_table("work_queue", schema=None) as batch_op:
        batch_op.drop_constraint("uq_work_queue__work_pool_id_name")
        batch_op.create_unique_constraint("uq_work_queue__name", ["name"])

    op.execute("DROP INDEX IF EXISTS ix_flow_run__work_queue_id_work_queue_name")
    op.execute("DROP INDEX IF EXISTS ix_deployment__work_queue_id_work_queue_name")



from copy import deepcopy
from typing import Any

from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi

from prefect import __version__ as PREFECT_VERSION


def inject_schemas_into_openapi(
    webserver: FastAPI, schemas_to_inject: dict[str, Any]
) -> dict[str, Any]:
    """
    Augments the webserver's OpenAPI schema with additional schemas from deployments / flows / tasks.

    Args:
        webserver: The FastAPI instance representing the webserver.
        schemas_to_inject: A dictionary of OpenAPI schemas to integrate.

    Returns:
        The augmented OpenAPI schema dictionary.
    """
    openapi_schema = get_openapi(
        title="FastAPI Prefect Runner", version=PREFECT_VERSION, routes=webserver.routes
    )

    augmented_schema = merge_definitions(schemas_to_inject, openapi_schema)
    return update_refs_to_components(augmented_schema)


def merge_definitions(
    injected_schemas: dict[str, Any], openapi_schema: dict[str, Any]
) -> dict[str, Any]:
    """
    Integrates definitions from injected schemas into the OpenAPI components.

    Args:
        injected_schemas: A dictionary of deployment-specific schemas.
        openapi_schema: The base OpenAPI schema to update.
    """
    openapi_schema_copy = deepcopy(openapi_schema)
    components = openapi_schema_copy.setdefault("components", {}).setdefault(
        "schemas", {}
    )
    for definitions in injected_schemas.values():
        if "definitions" in definitions:
            for def_name, def_schema in definitions["definitions"].items():
                def_schema_copy = deepcopy(def_schema)
                update_refs_in_schema(def_schema_copy, "#/components/schemas/")
                components[def_name] = def_schema_copy
    return openapi_schema_copy


def update_refs_in_schema(schema_item: Any, new_ref: str) -> None:
    """
    Recursively replaces `$ref` with a new reference base in a schema item.

    Args:
        schema_item: A schema or part of a schema to update references in.
        new_ref: The new base string to replace in `$ref` values.
    """
    if isinstance(schema_item, dict):
        if "$ref" in schema_item:
            schema_item["$ref"] = schema_item["$ref"].replace("#/definitions/", new_ref)
        for value in schema_item.values():
            update_refs_in_schema(value, new_ref)
    elif isinstance(schema_item, list):
        for item in schema_item:
            update_refs_in_schema(item, new_ref)


def update_refs_to_components(openapi_schema: dict[str, Any]) -> dict[str, Any]:
    """
    Updates all `$ref` fields in the OpenAPI schema to reference the components section.

    Args:
        openapi_schema: The OpenAPI schema to modify `$ref` fields in.
    """
    for path_item in openapi_schema.get("paths", {}).values():
        for operation in path_item.values():
            schema = (
                operation.get("requestBody", {})
                .get("content", {})
                .get("application/json", {})
                .get("schema", {})
            )
            update_refs_in_schema(schema, "#/components/schemas/")

    for definition in openapi_schema.get("definitions", {}).values():
        update_refs_in_schema(definition, "#/components/schemas/")

    return openapi_schema



"""json_variables

Revision ID: 94622c1663e8
Revises: b23c83a12cb4
Create Date: 2024-05-21 10:14:57.246286

"""

import json

import sqlalchemy as sa
from alembic import op

from prefect.server.utilities.database import JSON

# revision identifiers, used by Alembic.
revision = "94622c1663e8"
down_revision = "b23c83a12cb4"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("variable", sa.Column("json_value", JSON, nullable=True))

    conn = op.get_bind()

    result = conn.execute(sa.text("SELECT id, value FROM variable"))
    rows = result.fetchall()

    for variable_id, value in rows:
        # these values need to be json compatible strings
        json_value = json.dumps(value)
        conn.execute(
            sa.text("UPDATE variable SET json_value = :json_value WHERE id = :id"),
            {"json_value": json_value, "id": variable_id},
        )

    op.drop_column("variable", "value")
    op.alter_column("variable", "json_value", new_column_name="value")


def downgrade():
    op.add_column("variable", sa.Column("string_value", sa.String, nullable=True))

    conn = op.get_bind()

    result = conn.execute(sa.text("SELECT id, value FROM variable"))
    rows = result.fetchall()

    for variable_id, value in rows:
        string_value = str(value)
        conn.execute(
            sa.text("UPDATE variable SET string_value = :string_value WHERE id = :id"),
            {"string_value": string_value, "id": variable_id},
        )

    op.drop_column("variable", "value")
    op.alter_column("variable", "string_value", new_column_name="value", nullable=False)



"""
Functions for interacting with deployment ORM objects.
Intended for internal use by the Prefect REST API.
"""

import datetime
from typing import Dict, Iterable, List, Optional, Sequence, TypeVar, cast
from uuid import UUID, uuid4

import pendulum
import sqlalchemy as sa
from sqlalchemy import delete, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import Select

from prefect.server import models, schemas
from prefect.server.database import orm_models
from prefect.server.database.dependencies import db_injector
from prefect.server.database.interface import PrefectDBInterface
from prefect.server.events.clients import PrefectServerEventsClient
from prefect.server.exceptions import ObjectNotFoundError
from prefect.server.models.events import deployment_status_event
from prefect.server.schemas.statuses import DeploymentStatus
from prefect.server.utilities.database import json_contains
from prefect.settings import (
    PREFECT_API_SERVICES_SCHEDULER_MAX_RUNS,
    PREFECT_API_SERVICES_SCHEDULER_MAX_SCHEDULED_TIME,
    PREFECT_API_SERVICES_SCHEDULER_MIN_RUNS,
    PREFECT_API_SERVICES_SCHEDULER_MIN_SCHEDULED_TIME,
)

T = TypeVar("T", bound=tuple)


async def _delete_scheduled_runs(
    session: AsyncSession,
    deployment_id: UUID,
    auto_scheduled_only: bool = False,
) -> None:
    """
    This utility function deletes all of a deployment's runs that are in a Scheduled state
    and haven't run yet. It should be run any time a deployment is created or
    modified in order to ensure that future runs comply with the deployment's latest values.

    Args:
        deployment_id: the deployment for which we should delete runs.
        auto_scheduled_only: if True, only delete auto scheduled runs. Defaults to `False`.
    """
    delete_query = sa.delete(orm_models.FlowRun).where(
        orm_models.FlowRun.deployment_id == deployment_id,
        orm_models.FlowRun.state_type == schemas.states.StateType.SCHEDULED.value,
        orm_models.FlowRun.run_count == 0,
    )

    if auto_scheduled_only:
        delete_query = delete_query.where(
            orm_models.FlowRun.auto_scheduled.is_(True),
        )

    await session.execute(delete_query)


@db_injector
async def create_deployment(
    db: PrefectDBInterface,
    session: AsyncSession,
    deployment: schemas.core.Deployment,
) -> Optional[orm_models.Deployment]:
    """Upserts a deployment.

    Args:
        session: a database session
        deployment: a deployment model

    Returns:
        orm_models.Deployment: the newly-created or updated deployment

    """

    # set `updated` manually
    # known limitation of `on_conflict_do_update`, will not use `Column.onupdate`
    # https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#the-set-clause
    deployment.updated = pendulum.now("UTC")  # type: ignore[assignment]

    deployment.labels = await with_system_labels_for_deployment(session, deployment)

    schedules = deployment.schedules
    insert_values = deployment.model_dump_for_orm(
        exclude_unset=True, exclude={"schedules"}
    )

    requested_concurrency_limit = insert_values.pop("concurrency_limit", "unset")

    # The job_variables field in client and server schemas is named
    # infra_overrides in the database.
    job_variables = insert_values.pop("job_variables", None)
    if job_variables:
        insert_values["infra_overrides"] = job_variables

    conflict_update_fields = deployment.model_dump_for_orm(
        exclude_unset=True,
        exclude={
            "id",
            "created",
            "created_by",
            "schedules",
            "job_variables",
            "concurrency_limit",
        },
    )
    if job_variables:
        conflict_update_fields["infra_overrides"] = job_variables

    insert_stmt = (
        db.insert(orm_models.Deployment)
        .values(**insert_values)
        .on_conflict_do_update(
            index_elements=db.deployment_unique_upsert_columns,
            set_={**conflict_update_fields},
        )
    )

    await session.execute(insert_stmt)

    # Get the id of the deployment we just created or updated
    result = await session.execute(
        sa.select(orm_models.Deployment.id).where(
            sa.and_(
                orm_models.Deployment.flow_id == deployment.flow_id,
                orm_models.Deployment.name == deployment.name,
            )
        )
    )
    deployment_id = result.scalar_one_or_none()

    if not deployment_id:
        return None

    # Because this was possibly an upsert, we need to delete any existing
    # schedules and any runs from the old deployment.

    await _delete_scheduled_runs(
        session=session, deployment_id=deployment_id, auto_scheduled_only=True
    )

    await delete_schedules_for_deployment(session=session, deployment_id=deployment_id)

    if schedules:
        await create_deployment_schedules(
            session=session,
            deployment_id=deployment_id,
            schedules=[
                schemas.actions.DeploymentScheduleCreate(
                    schedule=schedule.schedule,
                    active=schedule.active,  # type: ignore[call-arg]
                )
                for schedule in schedules
            ],
        )

    if requested_concurrency_limit != "unset":
        await _create_or_update_deployment_concurrency_limit(
            session, deployment_id, deployment.concurrency_limit
        )

    query = (
        sa.select(orm_models.Deployment)
        .where(
            sa.and_(
                orm_models.Deployment.flow_id == deployment.flow_id,
                orm_models.Deployment.name == deployment.name,
            )
        )
        .execution_options(populate_existing=True)
    )
    refreshed_result = await session.execute(query)
    return refreshed_result.scalar()


async def update_deployment(
    session: AsyncSession,
    deployment_id: UUID,
    deployment: schemas.actions.DeploymentUpdate,
) -> bool:
    """Updates a deployment.

    Args:
        session: a database session
        deployment_id: the ID of the deployment to modify
        deployment: changes to a deployment model

    Returns:
        bool: whether the deployment was updated

    """

    from prefect.server.api.workers import WorkerLookups

    schedules = deployment.schedules

    # exclude_unset=True allows us to only update values provided by
    # the user, ignoring any defaults on the model
    update_data = deployment.model_dump_for_orm(
        exclude_unset=True,
        exclude={"work_pool_name"},
    )

    requested_concurrency_limit_update = update_data.pop("concurrency_limit", "unset")

    # The job_variables field in client and server schemas is named
    # infra_overrides in the database.
    job_variables = update_data.pop("job_variables", None)
    if job_variables:
        update_data["infra_overrides"] = job_variables

    should_update_schedules = update_data.pop("schedules", None) is not None

    if deployment.work_pool_name and deployment.work_queue_name:
        # If a specific pool name/queue name combination was provided, get the
        # ID for that work pool queue.
        update_data[
            "work_queue_id"
        ] = await WorkerLookups()._get_work_queue_id_from_name(
            session=session,
            work_pool_name=deployment.work_pool_name,
            work_queue_name=deployment.work_queue_name,
            create_queue_if_not_found=True,
        )
    elif deployment.work_pool_name:
        # If just a pool name was provided, get the ID for its default
        # work pool queue.
        update_data[
            "work_queue_id"
        ] = await WorkerLookups()._get_default_work_queue_id_from_work_pool_name(
            session=session,
            work_pool_name=deployment.work_pool_name,
        )
    elif deployment.work_queue_name:
        # If just a queue name was provided, ensure the queue exists and
        # get its ID.
        work_queue = await models.work_queues.ensure_work_queue_exists(
            session=session, name=update_data["work_queue_name"]
        )
        update_data["work_queue_id"] = work_queue.id

    update_stmt = (
        sa.update(orm_models.Deployment)
        .where(orm_models.Deployment.id == deployment_id)
        .values(**update_data)
    )
    result = await session.execute(update_stmt)

    # delete any auto scheduled runs that would have reflected the old deployment config
    await _delete_scheduled_runs(
        session=session, deployment_id=deployment_id, auto_scheduled_only=True
    )

    if should_update_schedules:
        # If schedules were provided, remove the existing schedules and
        # replace them with the new ones.
        await delete_schedules_for_deployment(
            session=session, deployment_id=deployment_id
        )
        await create_deployment_schedules(
            session=session,
            deployment_id=deployment_id,
            schedules=[
                schemas.actions.DeploymentScheduleCreate(
                    schedule=schedule.schedule,
                    active=schedule.active,  # type: ignore[call-arg]
                )
                for schedule in schedules
            ],
        )

    if requested_concurrency_limit_update != "unset":
        await _create_or_update_deployment_concurrency_limit(
            session, deployment_id, deployment.concurrency_limit
        )

    return result.rowcount > 0


async def _create_or_update_deployment_concurrency_limit(
    session: AsyncSession, deployment_id: UUID, limit: Optional[int]
):
    deployment = await session.get(orm_models.Deployment, deployment_id)
    assert deployment is not None

    if (
        deployment.global_concurrency_limit
        and deployment.global_concurrency_limit.limit == limit
    ) or (deployment.global_concurrency_limit is None and limit is None):
        return

    deployment._concurrency_limit = limit
    if limit is None:
        await _delete_related_concurrency_limit(
            session=session, deployment_id=deployment_id
        )
        await session.refresh(deployment)
    elif deployment.global_concurrency_limit:
        deployment.global_concurrency_limit.limit = limit
    else:
        limit_name = f"deployment:{deployment_id}"
        new_limit = orm_models.ConcurrencyLimitV2(name=limit_name, limit=limit)
        deployment.global_concurrency_limit = new_limit

    session.add(deployment)


async def read_deployment(
    session: AsyncSession, deployment_id: UUID
) -> Optional[orm_models.Deployment]:
    """Reads a deployment by id.

    Args:
        session: A database session
        deployment_id: a deployment id

    Returns:
        orm_models.Deployment: the deployment
    """

    return await session.get(orm_models.Deployment, deployment_id)


async def read_deployment_by_name(
    session: AsyncSession, name: str, flow_name: str
) -> Optional[orm_models.Deployment]:
    """Reads a deployment by name.

    Args:
        session: A database session
        name: a deployment name
        flow_name: the name of the flow the deployment belongs to

    Returns:
        orm_models.Deployment: the deployment
    """

    result = await session.execute(
        select(orm_models.Deployment)
        .join(orm_models.Flow, orm_models.Deployment.flow_id == orm_models.Flow.id)
        .where(
            sa.and_(
                orm_models.Flow.name == flow_name,
                orm_models.Deployment.name == name,
            )
        )
        .limit(1)
    )
    return result.scalar()


async def _apply_deployment_filters(
    query: Select[T],
    flow_filter: Optional[schemas.filters.FlowFilter] = None,
    flow_run_filter: Optional[schemas.filters.FlowRunFilter] = None,
    task_run_filter: Optional[schemas.filters.TaskRunFilter] = None,
    deployment_filter: Optional[schemas.filters.DeploymentFilter] = None,
    work_pool_filter: Optional[schemas.filters.WorkPoolFilter] = None,
    work_queue_filter: Optional[schemas.filters.WorkQueueFilter] = None,
) -> Select[T]:
    """
    Applies filters to a deployment query as a combination of EXISTS subqueries.
    """

    if deployment_filter:
        query = query.where(deployment_filter.as_sql_filter())

    if flow_filter:
        flow_exists_clause = select(orm_models.Deployment.id).where(
            orm_models.Deployment.flow_id == orm_models.Flow.id,
            flow_filter.as_sql_filter(),
        )

        query = query.where(flow_exists_clause.exists())

    if flow_run_filter or task_run_filter:
        flow_run_exists_clause = select(orm_models.FlowRun).where(
            orm_models.Deployment.id == orm_models.FlowRun.deployment_id
        )

        if flow_run_filter:
            flow_run_exists_clause = flow_run_exists_clause.where(
                flow_run_filter.as_sql_filter()
            )
        if task_run_filter:
            flow_run_exists_clause = flow_run_exists_clause.join(
                orm_models.TaskRun,
                orm_models.TaskRun.flow_run_id == orm_models.FlowRun.id,
            ).where(task_run_filter.as_sql_filter())

        query = query.where(flow_run_exists_clause.exists())

    if work_pool_filter or work_queue_filter:
        work_pool_exists_clause = select(orm_models.WorkQueue).where(
            orm_models.Deployment.work_queue_id == orm_models.WorkQueue.id
        )

        if work_queue_filter:
            work_pool_exists_clause = work_pool_exists_clause.where(
                work_queue_filter.as_sql_filter()
            )

        if work_pool_filter:
            work_pool_exists_clause = work_pool_exists_clause.join(
                orm_models.WorkPool,
                orm_models.WorkPool.id == orm_models.WorkQueue.work_pool_id,
            ).where(work_pool_filter.as_sql_filter())

        query = query.where(work_pool_exists_clause.exists())

    return query


async def read_deployments(
    session: AsyncSession,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
    flow_filter: Optional[schemas.filters.FlowFilter] = None,
    flow_run_filter: Optional[schemas.filters.FlowRunFilter] = None,
    task_run_filter: Optional[schemas.filters.TaskRunFilter] = None,
    deployment_filter: Optional[schemas.filters.DeploymentFilter] = None,
    work_pool_filter: Optional[schemas.filters.WorkPoolFilter] = None,
    work_queue_filter: Optional[schemas.filters.WorkQueueFilter] = None,
    sort: schemas.sorting.DeploymentSort = schemas.sorting.DeploymentSort.NAME_ASC,
) -> Sequence[orm_models.Deployment]:
    """
    Read deployments.

    Args:
        session: A database session
        offset: Query offset
        limit: Query limit
        flow_filter: only select deployments whose flows match these criteria
        flow_run_filter: only select deployments whose flow runs match these criteria
        task_run_filter: only select deployments whose task runs match these criteria
        deployment_filter: only select deployment that match these filters
        work_pool_filter: only select deployments whose work pools match these criteria
        work_queue_filter: only select deployments whose work pool queues match these criteria
        sort: the sort criteria for selected deployments. Defaults to `name` ASC.

    Returns:
        List[orm_models.Deployment]: deployments
    """

    query = select(orm_models.Deployment).order_by(sort.as_sql_sort())

    query = await _apply_deployment_filters(
        query=query,
        flow_filter=flow_filter,
        flow_run_filter=flow_run_filter,
        task_run_filter=task_run_filter,
        deployment_filter=deployment_filter,
        work_pool_filter=work_pool_filter,
        work_queue_filter=work_queue_filter,
    )

    if offset is not None:
        query = query.offset(offset)
    if limit is not None:
        query = query.limit(limit)

    result = await session.execute(query)
    return result.scalars().unique().all()


async def count_deployments(
    session: AsyncSession,
    flow_filter: Optional[schemas.filters.FlowFilter] = None,
    flow_run_filter: Optional[schemas.filters.FlowRunFilter] = None,
    task_run_filter: Optional[schemas.filters.TaskRunFilter] = None,
    deployment_filter: Optional[schemas.filters.DeploymentFilter] = None,
    work_pool_filter: Optional[schemas.filters.WorkPoolFilter] = None,
    work_queue_filter: Optional[schemas.filters.WorkQueueFilter] = None,
) -> int:
    """
    Count deployments.

    Args:
        session: A database session
        flow_filter: only count deployments whose flows match these criteria
        flow_run_filter: only count deployments whose flow runs match these criteria
        task_run_filter: only count deployments whose task runs match these criteria
        deployment_filter: only count deployment that match these filters
        work_pool_filter: only count deployments that match these work pool filters
        work_queue_filter: only count deployments that match these work pool queue filters

    Returns:
        int: the number of deployments matching filters
    """

    query = select(sa.func.count(sa.text("*"))).select_from(orm_models.Deployment)

    query = await _apply_deployment_filters(
        query=query,
        flow_filter=flow_filter,
        flow_run_filter=flow_run_filter,
        task_run_filter=task_run_filter,
        deployment_filter=deployment_filter,
        work_pool_filter=work_pool_filter,
        work_queue_filter=work_queue_filter,
    )

    result = await session.execute(query)
    return result.scalar_one()


async def delete_deployment(session: AsyncSession, deployment_id: UUID) -> bool:
    """
    Delete a deployment by id.

    Args:
        session: A database session
        deployment_id: a deployment id

    Returns:
        bool: whether or not the deployment was deleted
    """

    # delete scheduled runs, both auto- and user- created.
    await _delete_scheduled_runs(
        session=session, deployment_id=deployment_id, auto_scheduled_only=False
    )

    await _delete_related_concurrency_limit(
        session=session, deployment_id=deployment_id
    )

    result = await session.execute(
        delete(orm_models.Deployment).where(orm_models.Deployment.id == deployment_id)
    )
    return result.rowcount > 0


async def _delete_related_concurrency_limit(session: AsyncSession, deployment_id: UUID):
    return await session.execute(
        delete(orm_models.ConcurrencyLimitV2).where(
            orm_models.ConcurrencyLimitV2.id
            == sa.select(orm_models.Deployment.concurrency_limit_id)
            .where(orm_models.Deployment.id == deployment_id)
            .scalar_subquery()
        )
    )


async def schedule_runs(
    session: AsyncSession,
    deployment_id: UUID,
    start_time: Optional[datetime.datetime] = None,
    end_time: Optional[datetime.datetime] = None,
    min_time: Optional[datetime.timedelta] = None,
    min_runs: Optional[int] = None,
    max_runs: Optional[int] = None,
    auto_scheduled: bool = True,
) -> Sequence[UUID]:
    """
    Schedule flow runs for a deployment

    Args:
        session: a database session
        deployment_id: the id of the deployment to schedule
        start_time: the time from which to start scheduling runs
        end_time: runs will be scheduled until at most this time
        min_time: runs will be scheduled until at least this far in the future
        min_runs: a minimum amount of runs to schedule
        max_runs: a maximum amount of runs to schedule

    This function will generate the minimum number of runs that satisfy the min
    and max times, and the min and max counts. Specifically, the following order
    will be respected.

        - Runs will be generated starting on or after the `start_time`
        - No more than `max_runs` runs will be generated
        - No runs will be generated after `end_time` is reached
        - At least `min_runs` runs will be generated
        - Runs will be generated until at least `start_time` + `min_time` is reached

    Returns:
        a list of flow run ids scheduled for the deployment
    """
    if min_runs is None:
        min_runs = PREFECT_API_SERVICES_SCHEDULER_MIN_RUNS.value()
    if max_runs is None:
        max_runs = PREFECT_API_SERVICES_SCHEDULER_MAX_RUNS.value()
    if start_time is None:
        start_time = pendulum.now("UTC")
    if end_time is None:
        end_time = start_time + (
            PREFECT_API_SERVICES_SCHEDULER_MAX_SCHEDULED_TIME.value()
        )
    if min_time is None:
        min_time = PREFECT_API_SERVICES_SCHEDULER_MIN_SCHEDULED_TIME.value()

    actual_start_time = pendulum.instance(start_time)
    actual_end_time = pendulum.instance(end_time)

    runs = await _generate_scheduled_flow_runs(
        session=session,
        deployment_id=deployment_id,
        start_time=actual_start_time,
        end_time=actual_end_time,
        min_time=min_time,
        min_runs=min_runs,
        max_runs=max_runs,
        auto_scheduled=auto_scheduled,
    )
    return await _insert_scheduled_flow_runs(session=session, runs=runs)


async def _generate_scheduled_flow_runs(
    session: AsyncSession,
    deployment_id: UUID,
    start_time: datetime.datetime,
    end_time: datetime.datetime,
    min_time: datetime.timedelta,
    min_runs: int,
    max_runs: int,
    auto_scheduled: bool = True,
) -> List[Dict]:
    """
    Given a `deployment_id` and schedule, generates a list of flow run objects and
    associated scheduled states that represent scheduled flow runs. This method
    does NOT insert generated runs into the database, in order to facilitate
    batch operations. Call `_insert_scheduled_flow_runs()` to insert these runs.

    Runs include an idempotency key which prevents duplicate runs from being inserted
    if the output from this function is used more than once.

    Args:
        session: a database session
        deployment_id: the id of the deployment to schedule
        start_time: the time from which to start scheduling runs
        end_time: runs will be scheduled until at most this time
        min_time: runs will be scheduled until at least this far in the future
        min_runs: a minimum amount of runs to schedule
        max_runs: a maximum amount of runs to schedule

    This function will generate the minimum number of runs that satisfy the min
    and max times, and the min and max counts. Specifically, the following order
    will be respected.

        - Runs will be generated starting on or after the `start_time`
        - No more than `max_runs` runs will be generated
        - No runs will be generated after `end_time` is reached
        - At least `min_runs` runs will be generated
        - Runs will be generated until at least `start_time + min_time` is reached

    Returns:
        a list of dictionary representations of the `FlowRun` objects to schedule
    """
    runs = []

    deployment = await session.get(orm_models.Deployment, deployment_id)

    if not deployment:
        return []

    active_deployment_schedules = await read_deployment_schedules(
        session=session,
        deployment_id=deployment.id,
        deployment_schedule_filter=schemas.filters.DeploymentScheduleFilter(
            active=schemas.filters.DeploymentScheduleFilterActive(eq_=True)
        ),
    )

    for deployment_schedule in active_deployment_schedules:
        dates = []

        # generate up to `n` dates satisfying the min of `max_runs` and `end_time`
        for dt in deployment_schedule.schedule._get_dates_generator(
            n=max_runs, start=start_time, end=end_time
        ):
            dates.append(dt)

            # at any point, if we satisfy both of the minimums, we can stop
            if len(dates) >= min_runs and dt >= (start_time + min_time):
                break

        tags = deployment.tags
        if auto_scheduled:
            tags = ["auto-scheduled"] + tags

        for date in dates:
            runs.append(
                {
                    "id": uuid4(),
                    "flow_id": deployment.flow_id,
                    "deployment_id": deployment_id,
                    "deployment_version": deployment.version,
                    "work_queue_name": deployment.work_queue_name,
                    "work_queue_id": deployment.work_queue_id,
                    "parameters": deployment.parameters,
                    "infrastructure_document_id": deployment.infrastructure_document_id,
                    "idempotency_key": f"scheduled {deployment.id} {date}",
                    "tags": tags,
                    "auto_scheduled": auto_scheduled,
                    "state": schemas.states.Scheduled(
                        scheduled_time=date,
                        message="Flow run scheduled",
                    ).model_dump(),
                    "state_type": schemas.states.StateType.SCHEDULED,
                    "state_name": "Scheduled",
                    "next_scheduled_start_time": date,
                    "expected_start_time": date,
                }
            )

    return runs


@db_injector
async def _insert_scheduled_flow_runs(
    db: PrefectDBInterface, session: AsyncSession, runs: List[Dict]
) -> Sequence[UUID]:
    """
    Given a list of flow runs to schedule, as generated by `_generate_scheduled_flow_runs`,
    inserts them into the database. Note this is a separate method to facilitate batch
    operations on many scheduled runs.

    Args:
        session: a database session
        runs: a list of dicts representing flow runs to insert

    Returns:
        a list of flow run ids that were created
    """

    if not runs:
        return []

    # gracefully insert the flow runs against the idempotency key
    # this syntax (insert statement, values to insert) is most efficient
    # because it uses a single bind parameter
    await session.execute(
        db.insert(orm_models.FlowRun).on_conflict_do_nothing(
            index_elements=db.flow_run_unique_upsert_columns
        ),
        runs,
    )

    # query for the rows that were newly inserted (by checking for any flow runs with
    # no corresponding flow run states)
    inserted_rows = sa.select(db.FlowRun.id).where(
        db.FlowRun.id.in_([r["id"] for r in runs]),
        ~select(db.FlowRunState.id)
        .where(db.FlowRunState.flow_run_id == db.FlowRun.id)
        .exists(),
    )
    inserted_flow_run_ids = (await session.execute(inserted_rows)).scalars().all()

    # insert flow run states that correspond to the newly-insert rows
    insert_flow_run_states = [
        {"id": uuid4(), "flow_run_id": r["id"], **r["state"]}
        for r in runs
        if r["id"] in inserted_flow_run_ids
    ]
    if insert_flow_run_states:
        # this syntax (insert statement, values to insert) is most efficient
        # because it uses a single bind parameter
        await session.execute(
            orm_models.FlowRunState.__table__.insert(),  # type: ignore[attr-defined]
            insert_flow_run_states,
        )

        # set the `state_id` on the newly inserted runs
        stmt = db.set_state_id_on_inserted_flow_runs_statement(
            inserted_flow_run_ids=inserted_flow_run_ids,
            insert_flow_run_states=insert_flow_run_states,
        )

        await session.execute(stmt)

    return inserted_flow_run_ids


async def check_work_queues_for_deployment(
    session: AsyncSession, deployment_id: UUID
) -> Sequence[orm_models.WorkQueue]:
    """
    Get work queues that can pick up the specified deployment.

    Work queues will pick up a deployment when all of the following are met.

    - The deployment has ALL tags that the work queue has (i.e. the work
    queue's tags must be a subset of the deployment's tags).
    - The work queue's specified deployment IDs match the deployment's ID,
    or the work queue does NOT have specified deployment IDs.
    - The work queue's specified flow runners match the deployment's flow
    runner or the work queue does NOT have a specified flow runner.

    Notes on the query:

    - Our database currently allows either "null" and empty lists as
    null values in filters, so we need to catch both cases with "or".
    - `json_contains(A, B)` should be interpreted as "True if A
    contains B".

    Returns:
        List[orm_models.WorkQueue]: WorkQueues
    """
    deployment = await session.get(orm_models.Deployment, deployment_id)
    if not deployment:
        raise ObjectNotFoundError(f"Deployment with id {deployment_id} not found")

    query = (
        select(orm_models.WorkQueue)
        # work queue tags are a subset of deployment tags
        .filter(
            or_(
                json_contains(deployment.tags, orm_models.WorkQueue.filter["tags"]),
                json_contains([], orm_models.WorkQueue.filter["tags"]),
                json_contains(None, orm_models.WorkQueue.filter["tags"]),
            )
        )
        # deployment_ids is null or contains the deployment's ID
        .filter(
            or_(
                json_contains(
                    orm_models.WorkQueue.filter["deployment_ids"],
                    str(deployment.id),
                ),
                json_contains(None, orm_models.WorkQueue.filter["deployment_ids"]),
                json_contains([], orm_models.WorkQueue.filter["deployment_ids"]),
            )
        )
    )

    result = await session.execute(query)
    return result.scalars().unique().all()


async def create_deployment_schedules(
    session: AsyncSession,
    deployment_id: UUID,
    schedules: List[schemas.actions.DeploymentScheduleCreate],
) -> List[schemas.core.DeploymentSchedule]:
    """
    Creates a deployment's schedules.

    Args:
        session: A database session
        deployment_id: a deployment id
        schedules: a list of deployment schedule create actions
    """

    schedules_with_deployment_id = []
    for schedule in schedules:
        data = schedule.model_dump()
        data["deployment_id"] = deployment_id
        schedules_with_deployment_id.append(data)

    models = [
        orm_models.DeploymentSchedule(**schedule)
        for schedule in schedules_with_deployment_id
    ]
    session.add_all(models)
    await session.flush()

    return [
        schemas.core.DeploymentSchedule.model_validate(m, from_attributes=True)
        for m in models
    ]


async def read_deployment_schedules(
    session: AsyncSession,
    deployment_id: UUID,
    deployment_schedule_filter: Optional[
        schemas.filters.DeploymentScheduleFilter
    ] = None,
) -> List[schemas.core.DeploymentSchedule]:
    """
    Reads a deployment's schedules.

    Args:
        session: A database session
        deployment_id: a deployment id

    Returns:
        list[schemas.core.DeploymentSchedule]: the deployment's schedules
    """

    query = (
        sa.select(orm_models.DeploymentSchedule)
        .where(orm_models.DeploymentSchedule.deployment_id == deployment_id)
        .order_by(orm_models.DeploymentSchedule.updated.desc())
    )

    if deployment_schedule_filter:
        query = query.where(deployment_schedule_filter.as_sql_filter())

    result = await session.execute(query)

    return [
        schemas.core.DeploymentSchedule.model_validate(s, from_attributes=True)
        for s in result.scalars().all()
    ]


async def update_deployment_schedule(
    session: AsyncSession,
    deployment_id: UUID,
    deployment_schedule_id: UUID,
    schedule: schemas.actions.DeploymentScheduleUpdate,
) -> bool:
    """
    Updates a deployment's schedules.

    Args:
        session: A database session
        deployment_schedule_id: a deployment schedule id
        schedule: a deployment schedule update action
    """

    result = await session.execute(
        sa.update(orm_models.DeploymentSchedule)
        .where(
            sa.and_(
                orm_models.DeploymentSchedule.id == deployment_schedule_id,
                orm_models.DeploymentSchedule.deployment_id == deployment_id,
            )
        )
        .values(**schedule.model_dump(exclude_none=True))
    )

    return result.rowcount > 0


async def delete_schedules_for_deployment(
    session: AsyncSession, deployment_id: UUID
) -> bool:
    """
    Deletes a deployment schedule.

    Args:
        session: A database session
        deployment_id: a deployment id
    """

    result = await session.execute(
        sa.delete(orm_models.DeploymentSchedule).where(
            orm_models.DeploymentSchedule.deployment_id == deployment_id
        )
    )

    return result.rowcount > 0


async def delete_deployment_schedule(
    session: AsyncSession,
    deployment_id: UUID,
    deployment_schedule_id: UUID,
) -> bool:
    """
    Deletes a deployment schedule.

    Args:
        session: A database session
        deployment_schedule_id: a deployment schedule id
    """

    result = await session.execute(
        sa.delete(orm_models.DeploymentSchedule).where(
            sa.and_(
                orm_models.DeploymentSchedule.id == deployment_schedule_id,
                orm_models.DeploymentSchedule.deployment_id == deployment_id,
            )
        )
    )

    return result.rowcount > 0


@db_injector
async def mark_deployments_ready(
    db: PrefectDBInterface,
    deployment_ids: Optional[Iterable[UUID]] = None,
    work_queue_ids: Optional[Iterable[UUID]] = None,
) -> None:
    deployment_ids = deployment_ids or []
    work_queue_ids = work_queue_ids or []

    if not deployment_ids and not work_queue_ids:
        return

    async with db.session_context(
        begin_transaction=True,
    ) as session:
        result = await session.execute(
            select(orm_models.Deployment.id).where(
                sa.or_(
                    orm_models.Deployment.id.in_(deployment_ids),
                    orm_models.Deployment.work_queue_id.in_(work_queue_ids),
                ),
                orm_models.Deployment.status == DeploymentStatus.NOT_READY,
            )
        )
        unready_deployments = list(result.scalars().unique().all())

        last_polled = pendulum.now("UTC")

        await session.execute(
            sa.update(orm_models.Deployment)
            .where(
                sa.or_(
                    orm_models.Deployment.id.in_(deployment_ids),
                    orm_models.Deployment.work_queue_id.in_(work_queue_ids),
                )
            )
            .values(status=DeploymentStatus.READY, last_polled=last_polled)
        )

        if not unready_deployments:
            return

        async with PrefectServerEventsClient() as events:
            for deployment_id in unready_deployments:
                await events.emit(
                    await deployment_status_event(
                        session=session,
                        deployment_id=deployment_id,
                        status=DeploymentStatus.READY,
                        occurred=last_polled,
                    )
                )


@db_injector
async def mark_deployments_not_ready(
    db: PrefectDBInterface,
    deployment_ids: Optional[Iterable[UUID]] = None,
    work_queue_ids: Optional[Iterable[UUID]] = None,
) -> None:
    deployment_ids = deployment_ids or []
    work_queue_ids = work_queue_ids or []

    if not deployment_ids and not work_queue_ids:
        return

    async with db.session_context(
        begin_transaction=True,
    ) as session:
        result = await session.execute(
            select(orm_models.Deployment.id).where(
                sa.or_(
                    orm_models.Deployment.id.in_(deployment_ids),
                    orm_models.Deployment.work_queue_id.in_(work_queue_ids),
                ),
                orm_models.Deployment.status == DeploymentStatus.READY,
            )
        )
        ready_deployments = list(result.scalars().unique().all())

        await session.execute(
            sa.update(orm_models.Deployment)
            .where(
                sa.or_(
                    orm_models.Deployment.id.in_(deployment_ids),
                    orm_models.Deployment.work_queue_id.in_(work_queue_ids),
                )
            )
            .values(status=DeploymentStatus.NOT_READY)
        )

        if not ready_deployments:
            return

        async with PrefectServerEventsClient() as events:
            for deployment_id in ready_deployments:
                await events.emit(
                    await deployment_status_event(
                        session=session,
                        deployment_id=deployment_id,
                        status=DeploymentStatus.NOT_READY,
                        occurred=pendulum.now("UTC"),
                    )
                )


async def with_system_labels_for_deployment(
    session: AsyncSession,
    deployment: schemas.core.Deployment,
) -> schemas.core.KeyValueLabels:
    """Augment user supplied labels with system default labels for a deployment."""
    default_labels = cast(
        schemas.core.KeyValueLabels,
        {
            "prefect.flow.id": str(deployment.flow_id),
        },
    )

    user_supplied_labels = deployment.labels or {}

    parent_labels = (
        await models.flows.read_flow_labels(session, deployment.flow_id)
    ) or {}

    return parent_labels | default_labels | user_supplied_labels



"""Expand deployment schema for improved ux

Revision ID: 296e2665785f
Revises: 53c19b31aa09
Create Date: 2022-08-06 13:00:09.419944

"""

import sqlalchemy as sa
from alembic import op

import prefect

# revision identifiers, used by Alembic.
revision = "296e2665785f"
down_revision = "53c19b31aa09"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("deployment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "infra_overrides",
                prefect.server.utilities.database.JSON(astext_type=sa.Text()),
                server_default="{}",
                nullable=False,
            )
        )
        batch_op.add_column(
            sa.Column(
                "path",
                sa.String(),
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column(
                "entrypoint",
                sa.String(),
                nullable=True,
            )
        )


def downgrade():
    with op.batch_alter_table("deployment", schema=None) as batch_op:
        batch_op.drop_column("entrypoint")
        batch_op.drop_column("path")
        batch_op.drop_column("infra_overrides")



"""Index and backfill block_type_name

Revision ID: 22ef3915ccd8
Revises: cef24af2ec34
Create Date: 2023-10-30 10:37:20.922002

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "22ef3915ccd8"
down_revision = "cef24af2ec34"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_block_document__block_type_name_name"),
            ["block_type_name", "name"],
            unique=False,
        )

    backfill_query = """
        UPDATE block_document
        SET block_type_name = (SELECT name from block_type where block_type.id = block_document.block_type_id)
        WHERE block_document.id in (SELECT id from block_document where block_type_name is null limit 500);
    """

    with op.get_context().autocommit_block():
        conn = op.get_bind()
        while True:
            # Execute until we've backfilled all block_type_names
            result = conn.execute(sa.text(backfill_query))
            if result.rowcount <= 0:
                break


def downgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
             DROP INDEX
             ix_block_document__block_type_name_name
             """
        )



"""
Functions for interacting with block document ORM objects.
Intended for internal use by the Prefect REST API.
"""

from copy import copy
from typing import Dict, List, Optional, Sequence, Tuple, TypeVar, Union
from uuid import UUID, uuid4

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import Select

import prefect.server.models as models
from prefect.server import schemas
from prefect.server.database import orm_models
from prefect.server.database.dependencies import db_injector
from prefect.server.database.interface import PrefectDBInterface
from prefect.server.schemas.actions import BlockDocumentReferenceCreate
from prefect.server.schemas.core import BlockDocument
from prefect.server.schemas.filters import BlockSchemaFilter
from prefect.server.utilities.database import UUID as UUIDTypeDecorator
from prefect.utilities.collections import dict_to_flatdict, flatdict_to_dict
from prefect.utilities.names import obfuscate

T = TypeVar("T", bound=tuple)


async def create_block_document(
    session: AsyncSession,
    block_document: schemas.actions.BlockDocumentCreate,
) -> BlockDocument:
    # lookup block type name and copy to the block document table
    block_type = await models.block_types.read_block_type(
        session=session, block_type_id=block_document.block_type_id
    )
    assert block_type, f"Block type {block_document.block_type_id} not found"

    name: Union[str, None]
    # anonymous block documents can be given a random name if none is provided
    if block_document.is_anonymous and not block_document.name:
        name = f"anonymous-{uuid4()}"
    else:
        name = block_document.name

    orm_block = orm_models.BlockDocument(
        name=name,
        block_schema_id=block_document.block_schema_id,
        block_type_id=block_document.block_type_id,
        block_type_name=block_type.name,
        is_anonymous=block_document.is_anonymous,
    )

    (
        block_document_data_without_refs,
        block_document_references,
    ) = _separate_block_references_from_data(block_document.data)

    # encrypt the data and store in block document
    await orm_block.encrypt_data(session=session, data=block_document_data_without_refs)

    session.add(orm_block)
    await session.flush()

    # Create a block document reference for each reference in the block document data
    for key, reference_block_document_id in block_document_references:
        await create_block_document_reference(
            session=session,
            block_document_reference=BlockDocumentReferenceCreate(
                parent_block_document_id=orm_block.id,
                reference_block_document_id=reference_block_document_id,
                name=key,
            ),
        )

    # reload the block document in order to load the associated block schema
    # relationship
    new_block_document = await read_block_document_by_id(
        session=session,
        block_document_id=orm_block.id,
        include_secrets=False,
    )
    assert new_block_document

    return new_block_document


async def block_document_with_unique_values_exists(
    session: AsyncSession, block_type_id: UUID, name: str
) -> bool:
    result = await session.execute(
        sa.select(sa.exists(orm_models.BlockDocument)).where(
            orm_models.BlockDocument.block_type_id == block_type_id,
            orm_models.BlockDocument.name == name,
        )
    )
    return bool(result.scalar_one_or_none())


def _separate_block_references_from_data(
    block_document_data: Dict,
) -> Tuple[Dict, List[Tuple[str, UUID]]]:
    """
    Separates block document references from block document data so that a block
    document can be saved without references and the corresponding block document
    references can be saved.

    Args:
        block_document_data: Dictionary of block document data passed with request
            to create new block document

    Returns:
        block_document_data_with_out_refs: A copy of the block_document_data supplied
            with the block document references removed.
        block_document_references: A list of tuples each containing the name of the
            field referencing a block document and the ID of the referenced block
            document,
    """
    block_document_references = []
    block_document_data_without_refs = {}
    for key, value in block_document_data.items():
        # Current assumption is that any block references will be stored on a key of
        # the block document data and not nested in any other data structures.
        if isinstance(value, dict) and "$ref" in value:
            reference_block_document_id = value["$ref"].get("block_document_id")
            if reference_block_document_id is None:
                raise ValueError(
                    f"Received block reference without a block_document_id in key {key}"
                )
            block_document_references.append((key, reference_block_document_id))
        else:
            block_document_data_without_refs[key] = value
    return block_document_data_without_refs, block_document_references


async def read_block_document_by_id(
    session: AsyncSession,
    block_document_id: UUID,
    include_secrets: bool = False,
) -> Union[BlockDocument, None]:
    block_documents = await read_block_documents(
        session=session,
        block_document_filter=schemas.filters.BlockDocumentFilter(
            id=dict(any_=[block_document_id]),
            # don't apply any anonymous filtering
            is_anonymous=None,
        ),
        include_secrets=include_secrets,
        limit=1,
    )
    return block_documents[0] if block_documents else None


async def _construct_full_block_document(
    session: AsyncSession,
    block_documents_with_references: Sequence[
        Tuple[orm_models.ORMBlockDocument, Optional[str], Optional[UUID]]
    ],
    parent_block_document: Optional[BlockDocument] = None,
    include_secrets: bool = False,
) -> Optional[BlockDocument]:
    if len(block_documents_with_references) == 0:
        return None
    if parent_block_document is None:
        parent_block_document = copy(
            await _find_parent_block_document(
                session,
                block_documents_with_references,
                include_secrets=include_secrets,
            )
        )

    if parent_block_document is None:
        raise ValueError("Unable to determine parent block document")

    # Recursively walk block document tree and construct the full block
    # document data for each child and add it to the parent's block document
    # data
    for (
        orm_block_document,
        name,
        parent_block_document_id,
    ) in block_documents_with_references:
        if parent_block_document_id == parent_block_document.id and name is not None:
            block_document = await BlockDocument.from_orm_model(
                session, orm_block_document, include_secrets=include_secrets
            )
            full_child_block_document = await _construct_full_block_document(
                session,
                block_documents_with_references,
                parent_block_document=copy(block_document),
                include_secrets=include_secrets,
            )
            assert full_child_block_document
            parent_block_document.data[name] = full_child_block_document.data
            parent_block_document.block_document_references[name] = {
                "block_document": {
                    "id": block_document.id,
                    "name": block_document.name,
                    "block_type": block_document.block_type,
                    "is_anonymous": block_document.is_anonymous,
                    "block_document_references": (
                        full_child_block_document.block_document_references
                    ),
                }
            }

    return parent_block_document


async def _find_parent_block_document(
    session: AsyncSession,
    block_documents_with_references: Sequence[
        Tuple[orm_models.ORMBlockDocument, Optional[str], Optional[UUID]]
    ],
    include_secrets: bool = False,
) -> Union[BlockDocument, None]:
    parent_orm_block_document = next(
        (
            block_document
            for (
                block_document,
                _,
                parent_block_document_id,
            ) in block_documents_with_references
            if parent_block_document_id is None
        ),
        None,
    )
    return (
        await BlockDocument.from_orm_model(
            session,
            parent_orm_block_document,
            include_secrets=include_secrets,
        )
        if parent_orm_block_document is not None
        else None
    )


async def read_block_document_by_name(
    session: AsyncSession,
    name: str,
    block_type_slug: str,
    include_secrets: bool = False,
) -> Union[BlockDocument, None]:
    """
    Read a block document with the given name and block type slug.
    """
    block_documents = await read_block_documents(
        session=session,
        block_document_filter=schemas.filters.BlockDocumentFilter(
            name=dict(any_=[name]),
            # don't apply any anonymous filtering
            is_anonymous=None,
        ),
        block_type_filter=schemas.filters.BlockTypeFilter(
            slug=dict(any_=[block_type_slug])
        ),
        include_secrets=include_secrets,
        limit=1,
    )
    return block_documents[0] if block_documents else None


def _apply_block_document_filters(
    query: Select[T],
    block_document_filter: Optional[schemas.filters.BlockDocumentFilter] = None,
    block_schema_filter: Optional[schemas.filters.BlockSchemaFilter] = None,
    block_type_filter: Optional[schemas.filters.BlockTypeFilter] = None,
) -> Select[T]:
    # if no filter is provided, one is created that excludes anonymous blocks
    if block_document_filter is None:
        block_document_filter = schemas.filters.BlockDocumentFilter(
            is_anonymous=schemas.filters.BlockDocumentFilterIsAnonymous(eq_=False)
        )

    # --- Build an initial query that filters for the requested block documents
    query = query.where(block_document_filter.as_sql_filter())

    if block_type_filter is not None:
        block_type_exists_clause = sa.select(orm_models.BlockType).where(
            orm_models.BlockType.id == orm_models.BlockDocument.block_type_id,
            block_type_filter.as_sql_filter(),
        )
        query = query.where(block_type_exists_clause.exists())

    if block_schema_filter is not None:
        block_schema_exists_clause = sa.select(orm_models.BlockSchema).where(
            orm_models.BlockSchema.id == orm_models.BlockDocument.block_schema_id,
            block_schema_filter.as_sql_filter(),
        )
        query = query.where(block_schema_exists_clause.exists())

    return query


async def read_block_documents(
    session: AsyncSession,
    block_document_filter: Optional[schemas.filters.BlockDocumentFilter] = None,
    block_type_filter: Optional[schemas.filters.BlockTypeFilter] = None,
    block_schema_filter: Optional[schemas.filters.BlockSchemaFilter] = None,
    include_secrets: bool = False,
    sort: schemas.sorting.BlockDocumentSort = schemas.sorting.BlockDocumentSort.NAME_ASC,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> List[BlockDocument]:
    """
    Read block documents with an optional limit and offset
    """
    # --- Build an initial query that filters for the requested block documents
    filtered_block_documents_query = sa.select(orm_models.BlockDocument.id)
    filtered_block_documents_query = _apply_block_document_filters(
        query=filtered_block_documents_query,
        block_document_filter=block_document_filter,
        block_type_filter=block_type_filter,
        block_schema_filter=block_schema_filter,
    )
    filtered_block_documents_query = filtered_block_documents_query.order_by(
        sort.as_sql_sort()
    )

    if offset is not None:
        filtered_block_documents_query = filtered_block_documents_query.offset(offset)

    if limit is not None:
        filtered_block_documents_query = filtered_block_documents_query.limit(limit)

    filtered_block_documents_cte = filtered_block_documents_query.cte(
        "filtered_block_documents"
    )

    # --- Build a recursive query that starts with the filtered block documents
    # and iteratively loads all referenced block documents. The query includes
    # the ID of each block document as well as the ID of the document that
    # references it and name it's referenced by, if applicable.
    parent_documents = (
        sa.select(
            filtered_block_documents_cte.c.id,
            sa.cast(sa.null(), sa.String).label("reference_name"),
            sa.cast(sa.null(), UUIDTypeDecorator).label(
                "reference_parent_block_document_id"
            ),
        )
        .select_from(filtered_block_documents_cte)
        .cte("all_block_documents", recursive=True)
    )
    # recursive part of query
    referenced_documents = (
        sa.select(
            orm_models.BlockDocumentReference.reference_block_document_id,
            orm_models.BlockDocumentReference.name,
            orm_models.BlockDocumentReference.parent_block_document_id,
        )
        .select_from(parent_documents)
        .join(
            orm_models.BlockDocumentReference,
            orm_models.BlockDocumentReference.parent_block_document_id
            == parent_documents.c.id,
        )
    )
    # union the recursive CTE
    all_block_documents_query = parent_documents.union_all(referenced_documents)

    # --- Join the recursive query that contains all required document IDs
    # back to the BlockDocument table to load info for every document
    # and order by name
    final_query = (
        sa.select(
            orm_models.BlockDocument,
            all_block_documents_query.c.reference_name,
            all_block_documents_query.c.reference_parent_block_document_id,
        )
        .select_from(all_block_documents_query)
        .join(
            orm_models.BlockDocument,
            orm_models.BlockDocument.id == all_block_documents_query.c.id,
        )
        .order_by(sort.as_sql_sort())
    )

    result = await session.execute(
        final_query.execution_options(populate_existing=True)
    )

    block_documents_with_references = result.unique().all()

    # identify true "parent" documents as those with no reference parent ids
    parent_block_document_ids = [
        d[0].id
        for d in block_documents_with_references
        if d.reference_parent_block_document_id is None
    ]

    # walk the resulting dataset and hydrate all block documents
    fully_constructed_block_documents: List[BlockDocument] = []
    visited_block_document_ids = []
    for root_orm_block_document, _, _ in block_documents_with_references:
        if (
            root_orm_block_document.id in parent_block_document_ids
            and root_orm_block_document.id not in visited_block_document_ids
        ):
            root_block_document = await BlockDocument.from_orm_model(
                session, root_orm_block_document, include_secrets=include_secrets
            )
            constructed = await _construct_full_block_document(
                session,
                block_documents_with_references,  # type: ignore
                root_block_document,
                include_secrets=include_secrets,
            )
            assert constructed

            fully_constructed_block_documents.append(constructed)
            visited_block_document_ids.append(root_orm_block_document.id)

    block_schema_ids = [
        block_document.block_schema_id
        for block_document in fully_constructed_block_documents
    ]
    block_schemas = await models.block_schemas.read_block_schemas(
        session=session,
        block_schema_filter=BlockSchemaFilter(id=dict(any_=block_schema_ids)),
    )
    for block_document in fully_constructed_block_documents:
        corresponding_block_schema = next(
            block_schema
            for block_schema in block_schemas
            if block_schema.id == block_document.block_schema_id
        )
        block_document.block_schema = corresponding_block_schema

    return fully_constructed_block_documents


async def count_block_documents(
    session: AsyncSession,
    block_document_filter: Optional[schemas.filters.BlockDocumentFilter] = None,
    block_type_filter: Optional[schemas.filters.BlockTypeFilter] = None,
    block_schema_filter: Optional[schemas.filters.BlockSchemaFilter] = None,
) -> int:
    """
    Count block documents that match the filters.
    """
    query = sa.select(sa.func.count()).select_from(orm_models.BlockDocument)

    query = _apply_block_document_filters(
        query=query,
        block_document_filter=block_document_filter,
        block_schema_filter=block_schema_filter,
        block_type_filter=block_type_filter,
    )

    result = await session.execute(query)
    return result.scalar()  # type: ignore


async def delete_block_document(
    session: AsyncSession,
    block_document_id: UUID,
) -> bool:
    query = sa.delete(orm_models.BlockDocument).where(
        orm_models.BlockDocument.id == block_document_id
    )
    result = await session.execute(query)
    return result.rowcount > 0


async def update_block_document(
    session: AsyncSession,
    block_document_id: UUID,
    block_document: schemas.actions.BlockDocumentUpdate,
) -> bool:
    merge_existing_data = block_document.merge_existing_data
    current_block_document = await session.get(
        orm_models.BlockDocument, block_document_id
    )
    if not current_block_document:
        return False

    update_values = block_document.model_dump_for_orm(
        exclude_unset=merge_existing_data,
        exclude={"merge_existing_data"},
    )

    if "data" in update_values and update_values["data"] is not None:
        current_data = await current_block_document.decrypt_data(session=session)

        # if a value for a secret field was provided that is identical to the
        # obfuscated value of the current secret value, it means someone is
        # probably trying to update all of the documents fields without
        # realizing they are posting back obfuscated data, so we disregard the update
        flat_update_data = dict_to_flatdict(update_values["data"])
        flat_current_data = dict_to_flatdict(current_data)
        for secret_field in current_block_document.block_schema.fields.get(
            "secret_fields", []
        ):
            secret_key = tuple(secret_field.split("."))
            current_secret = flat_current_data.get(secret_key)
            if current_secret is not None:
                if flat_update_data.get(secret_key) == obfuscate(current_secret):
                    flat_update_data[secret_key] = current_secret
            # Looks for obfuscated values nested under a secret field with a wildcard.
            # If any obfuscated values are found, we assume that it shouldn't be update,
            # and they are replaced with the current value for that key to avoid losing
            # data during update.
            elif "*" in secret_key:
                wildcard_index = secret_key.index("*")
                for data_key in flat_update_data.keys():
                    if (
                        secret_key[0:wildcard_index] == data_key[0:wildcard_index]
                    ) and (
                        flat_update_data[data_key]
                        == obfuscate(flat_update_data[data_key])
                    ):
                        flat_update_data[data_key] = flat_current_data[data_key]

        update_values["data"] = flatdict_to_dict(flat_update_data)

        if merge_existing_data:
            # merge the existing data and the new data for partial updates
            current_data.update(update_values["data"])
            update_values["data"] = current_data

        current_block_document_references = (
            (
                await session.execute(
                    sa.select(orm_models.BlockDocumentReference).filter_by(
                        parent_block_document_id=block_document_id
                    )
                )
            )
            .scalars()
            .all()
        )
        (
            block_document_data_without_refs,
            new_block_document_references,
        ) = _separate_block_references_from_data(update_values["data"])

        # encrypt the data and write updated data to the block document
        await current_block_document.encrypt_data(
            session=session, data=block_document_data_without_refs
        )

        # `proposed_block_schema` is always the same as the schema on the client-side
        # Block class that is calling `save`, which may or may not be the same schema
        # as the one on the saved block document
        proposed_block_schema_id = block_document.block_schema_id

        # if a new schema is proposed, update the block schema id for the block document
        if (
            proposed_block_schema_id is not None
            and proposed_block_schema_id != current_block_document.block_schema_id
        ):
            proposed_block_schema = await session.get(
                orm_models.BlockSchema, proposed_block_schema_id
            )
            assert (
                proposed_block_schema
            ), f"Block schema {proposed_block_schema_id} not found"

            # make sure the proposed schema is of the same block type as the current document
            if (
                proposed_block_schema.block_type_id
                != current_block_document.block_type_id
            ):
                raise ValueError(
                    "Must migrate block document to a block schema of the same block"
                    " type."
                )
            await session.execute(
                sa.update(orm_models.BlockDocument)
                .where(orm_models.BlockDocument.id == block_document_id)
                .values(block_schema_id=proposed_block_schema_id)
            )

        unchanged_block_document_references = []
        for name, reference_block_document_id in new_block_document_references:
            matching_current_block_document_reference = _find_block_document_reference(
                current_block_document_references,
                name,
                reference_block_document_id,
            )
            if matching_current_block_document_reference is None:
                await create_block_document_reference(
                    session=session,
                    block_document_reference=BlockDocumentReferenceCreate(
                        parent_block_document_id=block_document_id,
                        reference_block_document_id=reference_block_document_id,
                        name=name,
                    ),
                )
            else:
                unchanged_block_document_references.append(
                    matching_current_block_document_reference
                )

        for block_document_reference in current_block_document_references:
            if block_document_reference not in unchanged_block_document_references:
                await delete_block_document_reference(
                    session, block_document_reference_id=block_document_reference.id
                )

    return True


def _find_block_document_reference(
    block_document_references: Sequence[orm_models.BlockDocumentReference],
    name: str,
    reference_block_document_id: UUID,
) -> Optional[orm_models.BlockDocumentReference]:
    return next(
        (
            block_document_reference
            for block_document_reference in block_document_references
            if block_document_reference.name == name
            and block_document_reference.reference_block_document_id
            == reference_block_document_id
        ),
        None,
    )


@db_injector
async def create_block_document_reference(
    db: PrefectDBInterface,
    session: AsyncSession,
    block_document_reference: schemas.actions.BlockDocumentReferenceCreate,
) -> Union[orm_models.BlockDocumentReference, None]:
    insert_stmt = db.insert(orm_models.BlockDocumentReference).values(
        **block_document_reference.model_dump_for_orm(
            exclude_unset=True, exclude={"created", "updated"}
        )
    )
    await session.execute(insert_stmt)

    result = await session.execute(
        sa.select(orm_models.BlockDocumentReference).where(
            orm_models.BlockDocumentReference.id == block_document_reference.id
        )
    )

    return result.scalar()


async def delete_block_document_reference(
    session: AsyncSession,
    block_document_reference_id: UUID,
) -> bool:
    query = sa.delete(orm_models.BlockDocumentReference).where(
        orm_models.BlockDocumentReference.id == block_document_reference_id
    )
    result = await session.execute(query)
    return result.rowcount > 0



import datetime
from typing import Any, ClassVar, Generic, Optional, TypeVar, Union
from uuid import UUID

from pydantic import ConfigDict, Field
from pydantic_extra_types.pendulum_dt import DateTime
from typing_extensions import Literal

import prefect.client.schemas.objects as objects
from prefect._internal.schemas.bases import ObjectBaseModel, PrefectBaseModel
from prefect._internal.schemas.fields import CreatedBy, UpdatedBy
from prefect.types import KeyValueLabelsField
from prefect.utilities.collections import AutoEnum
from prefect.utilities.names import generate_slug

T = TypeVar("T")


class SetStateStatus(AutoEnum):
    """Enumerates return statuses for setting run states."""

    ACCEPT = AutoEnum.auto()
    REJECT = AutoEnum.auto()
    ABORT = AutoEnum.auto()
    WAIT = AutoEnum.auto()


class StateAcceptDetails(PrefectBaseModel):
    """Details associated with an ACCEPT state transition."""

    type: Literal["accept_details"] = Field(
        default="accept_details",
        description=(
            "The type of state transition detail. Used to ensure pydantic does not"
            " coerce into a different type."
        ),
    )


class StateRejectDetails(PrefectBaseModel):
    """Details associated with a REJECT state transition."""

    type: Literal["reject_details"] = Field(
        default="reject_details",
        description=(
            "The type of state transition detail. Used to ensure pydantic does not"
            " coerce into a different type."
        ),
    )
    reason: Optional[str] = Field(
        default=None, description="The reason why the state transition was rejected."
    )


class StateAbortDetails(PrefectBaseModel):
    """Details associated with an ABORT state transition."""

    type: Literal["abort_details"] = Field(
        default="abort_details",
        description=(
            "The type of state transition detail. Used to ensure pydantic does not"
            " coerce into a different type."
        ),
    )
    reason: Optional[str] = Field(
        default=None, description="The reason why the state transition was aborted."
    )


class StateWaitDetails(PrefectBaseModel):
    """Details associated with a WAIT state transition."""

    type: Literal["wait_details"] = Field(
        default="wait_details",
        description=(
            "The type of state transition detail. Used to ensure pydantic does not"
            " coerce into a different type."
        ),
    )
    delay_seconds: int = Field(
        default=...,
        description=(
            "The length of time in seconds the client should wait before transitioning"
            " states."
        ),
    )
    reason: Optional[str] = Field(
        default=None, description="The reason why the state transition should wait."
    )


class HistoryResponseState(PrefectBaseModel):
    """Represents a single state's history over an interval."""

    state_type: objects.StateType = Field(default=..., description="The state type.")
    state_name: str = Field(default=..., description="The state name.")
    count_runs: int = Field(
        default=...,
        description="The number of runs in the specified state during the interval.",
    )
    sum_estimated_run_time: datetime.timedelta = Field(
        default=...,
        description="The total estimated run time of all runs during the interval.",
    )
    sum_estimated_lateness: datetime.timedelta = Field(
        default=...,
        description=(
            "The sum of differences between actual and expected start time during the"
            " interval."
        ),
    )


class HistoryResponse(PrefectBaseModel):
    """Represents a history of aggregation states over an interval"""

    interval_start: DateTime = Field(
        default=..., description="The start date of the interval."
    )
    interval_end: DateTime = Field(
        default=..., description="The end date of the interval."
    )
    states: list[HistoryResponseState] = Field(
        default=..., description="A list of state histories during the interval."
    )


StateResponseDetails = Union[
    StateAcceptDetails, StateWaitDetails, StateRejectDetails, StateAbortDetails
]


class OrchestrationResult(PrefectBaseModel, Generic[T]):
    """
    A container for the output of state orchestration.
    """

    state: Optional[objects.State[T]]
    status: SetStateStatus
    details: StateResponseDetails


class WorkerFlowRunResponse(PrefectBaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    work_pool_id: UUID
    work_queue_id: UUID
    flow_run: objects.FlowRun


class FlowRunResponse(ObjectBaseModel):
    name: str = Field(
        default_factory=lambda: generate_slug(2),
        description=(
            "The name of the flow run. Defaults to a random slug if not specified."
        ),
        examples=["my-flow-run"],
    )
    flow_id: UUID = Field(default=..., description="The id of the flow being run.")
    state_id: Optional[UUID] = Field(
        default=None, description="The id of the flow run's current state."
    )
    deployment_id: Optional[UUID] = Field(
        default=None,
        description=(
            "The id of the deployment associated with this flow run, if available."
        ),
    )
    deployment_version: Optional[str] = Field(
        default=None,
        description="The version of the deployment associated with this flow run.",
        examples=["1.0"],
    )
    work_queue_name: Optional[str] = Field(
        default=None, description="The work queue that handled this flow run."
    )
    flow_version: Optional[str] = Field(
        default=None,
        description="The version of the flow executed in this flow run.",
        examples=["1.0"],
    )
    parameters: dict[str, Any] = Field(
        default_factory=dict, description="Parameters for the flow run."
    )
    idempotency_key: Optional[str] = Field(
        default=None,
        description=(
            "An optional idempotency key for the flow run. Used to ensure the same flow"
            " run is not created multiple times."
        ),
    )
    context: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional context for the flow run.",
        examples=[{"my_var": "my_val"}],
    )
    empirical_policy: objects.FlowRunPolicy = Field(
        default_factory=objects.FlowRunPolicy,
    )
    tags: list[str] = Field(
        default_factory=list,
        description="A list of tags on the flow run",
        examples=[["tag-1", "tag-2"]],
    )
    labels: KeyValueLabelsField
    parent_task_run_id: Optional[UUID] = Field(
        default=None,
        description=(
            "If the flow run is a subflow, the id of the 'dummy' task in the parent"
            " flow used to track subflow state."
        ),
    )
    run_count: int = Field(
        default=0, description="The number of times the flow run was executed."
    )
    expected_start_time: Optional[DateTime] = Field(
        default=None,
        description="The flow run's expected start time.",
    )
    next_scheduled_start_time: Optional[DateTime] = Field(
        default=None,
        description="The next time the flow run is scheduled to start.",
    )
    start_time: Optional[DateTime] = Field(
        default=None, description="The actual start time."
    )
    end_time: Optional[DateTime] = Field(
        default=None, description="The actual end time."
    )
    total_run_time: datetime.timedelta = Field(
        default=datetime.timedelta(0),
        description=(
            "Total run time. If the flow run was executed multiple times, the time of"
            " each run will be summed."
        ),
    )
    estimated_run_time: datetime.timedelta = Field(
        default=datetime.timedelta(0),
        description="A real-time estimate of the total run time.",
    )
    estimated_start_time_delta: datetime.timedelta = Field(
        default=datetime.timedelta(0),
        description="The difference between actual and expected start time.",
    )
    auto_scheduled: bool = Field(
        default=False,
        description="Whether or not the flow run was automatically scheduled.",
    )
    infrastructure_document_id: Optional[UUID] = Field(
        default=None,
        description="The block document defining infrastructure to use this flow run.",
    )
    infrastructure_pid: Optional[str] = Field(
        default=None,
        description="The id of the flow run as returned by an infrastructure block.",
    )
    created_by: Optional[CreatedBy] = Field(
        default=None,
        description="Optional information about the creator of this flow run.",
    )
    work_queue_id: Optional[UUID] = Field(
        default=None, description="The id of the run's work pool queue."
    )

    work_pool_id: Optional[UUID] = Field(
        description="The work pool with which the queue is associated."
    )
    work_pool_name: Optional[str] = Field(
        default=None,
        description="The name of the flow run's work pool.",
        examples=["my-work-pool"],
    )
    state: Optional[objects.State] = Field(
        default=None,
        description="The state of the flow run.",
        examples=["objects.State(type=objects.StateType.COMPLETED)"],
    )
    job_variables: Optional[dict[str, Any]] = Field(
        default=None, description="Job variables for the flow run."
    )

    # These are server-side optimizations and should not be present on client models
    # TODO: Deprecate these fields

    state_type: Optional[objects.StateType] = Field(
        default=None, description="The type of the current flow run state."
    )
    state_name: Optional[str] = Field(
        default=None, description="The name of the current flow run state."
    )

    def __eq__(self, other: Any) -> bool:
        """
        Check for "equality" to another flow run schema

        Estimates times are rolling and will always change with repeated queries for
        a flow run so we ignore them during equality checks.
        """
        if isinstance(other, objects.FlowRun):
            exclude_fields = {"estimated_run_time", "estimated_start_time_delta"}
            return self.model_dump(exclude=exclude_fields) == other.model_dump(
                exclude=exclude_fields
            )
        return super().__eq__(other)


class DeploymentResponse(ObjectBaseModel):
    name: str = Field(default=..., description="The name of the deployment.")
    version: Optional[str] = Field(
        default=None, description="An optional version for the deployment."
    )
    description: Optional[str] = Field(
        default=None, description="A description for the deployment."
    )
    flow_id: UUID = Field(
        default=..., description="The flow id associated with the deployment."
    )
    concurrency_limit: Optional[int] = Field(
        default=None,
        description="DEPRECATED: Prefer `global_concurrency_limit`. Will always be None for backwards compatibility. Will be removed after December 2024.",
        deprecated=True,
    )
    global_concurrency_limit: Optional["GlobalConcurrencyLimitResponse"] = Field(
        default=None,
        description="The global concurrency limit object for enforcing the maximum number of flow runs that can be active at once.",
    )
    concurrency_options: Optional[objects.ConcurrencyOptions] = Field(
        default=None,
        description="The concurrency options for the deployment.",
    )
    paused: bool = Field(
        default=False, description="Whether or not the deployment is paused."
    )
    concurrency_options: Optional[objects.ConcurrencyOptions] = Field(
        default=None,
        description="The concurrency options for the deployment.",
    )
    schedules: list[objects.DeploymentSchedule] = Field(
        default_factory=list, description="A list of schedules for the deployment."
    )
    job_variables: dict[str, Any] = Field(
        default_factory=dict,
        description="Overrides to apply to flow run infrastructure at runtime.",
    )
    parameters: dict[str, Any] = Field(
        default_factory=dict,
        description="Parameters for flow runs scheduled by the deployment.",
    )
    pull_steps: Optional[list[dict[str, Any]]] = Field(
        default=None,
        description="Pull steps for cloning and running this deployment.",
    )
    tags: list[str] = Field(
        default_factory=list,
        description="A list of tags for the deployment",
        examples=[["tag-1", "tag-2"]],
    )
    labels: KeyValueLabelsField
    work_queue_name: Optional[str] = Field(
        default=None,
        description=(
            "The work queue for the deployment. If no work queue is set, work will not"
            " be scheduled."
        ),
    )
    last_polled: Optional[DateTime] = Field(
        default=None,
        description="The last time the deployment was polled for status updates.",
    )
    parameter_openapi_schema: Optional[dict[str, Any]] = Field(
        default=None,
        description="The parameter schema of the flow, including defaults.",
    )
    path: Optional[str] = Field(
        default=None,
        description=(
            "The path to the working directory for the workflow, relative to remote"
            " storage or an absolute path."
        ),
    )
    entrypoint: Optional[str] = Field(
        default=None,
        description=(
            "The path to the entrypoint for the workflow, relative to the `path`."
        ),
    )
    storage_document_id: Optional[UUID] = Field(
        default=None,
        description="The block document defining storage used for this flow.",
    )
    infrastructure_document_id: Optional[UUID] = Field(
        default=None,
        description="The block document defining infrastructure to use for flow runs.",
    )
    created_by: Optional[CreatedBy] = Field(
        default=None,
        description="Optional information about the creator of this deployment.",
    )
    updated_by: Optional[UpdatedBy] = Field(
        default=None,
        description="Optional information about the updater of this deployment.",
    )
    work_queue_id: Optional[UUID] = Field(
        default=None,
        description=(
            "The id of the work pool queue to which this deployment is assigned."
        ),
    )
    enforce_parameter_schema: bool = Field(
        default=True,
        description=(
            "Whether or not the deployment should enforce the parameter schema."
        ),
    )
    work_pool_name: Optional[str] = Field(
        default=None,
        description="The name of the deployment's work pool.",
    )
    status: Optional[objects.DeploymentStatus] = Field(
        default=None,
        description="Current status of the deployment.",
    )


class MinimalConcurrencyLimitResponse(PrefectBaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: UUID
    name: str
    limit: int


class GlobalConcurrencyLimitResponse(ObjectBaseModel):
    """
    A response object for global concurrency limits.
    """

    active: bool = Field(
        default=True, description="Whether the global concurrency limit is active."
    )
    name: str = Field(
        default=..., description="The name of the global concurrency limit."
    )
    limit: int = Field(default=..., description="The concurrency limit.")
    active_slots: int = Field(default=..., description="The number of active slots.")
    slot_decay_per_second: float = Field(
        default=2.0,
        description="The decay rate for active slots when used as a rate limit.",
    )



import pytest

from prefect import flow
from prefect.variables import Variable


@pytest.fixture
async def variable():
    var = await Variable.set(name="my_variable", value="my-value", tags=["123", "456"])
    return var


async def test_set_sync(variable):
    @flow
    def test_flow():
        # confirm variable doesn't exist
        variable_doesnt_exist = Variable.get("my_new_variable")
        assert variable_doesnt_exist is None

        # create new
        created = Variable.set(
            name="my_new_variable",
            value="my_value",
            tags=["123", "456"],
        )
        assert created.value == "my_value"
        assert created.tags == ["123", "456"]

        # try to overwrite
        with pytest.raises(
            ValueError,
            match="Variable 'my_new_variable' already exists. Use `overwrite=True` to update it.",
        ):
            Variable.set(name="my_new_variable", value="other_value")

        # actually overwrite
        updated = Variable.set(
            name="my_new_variable",
            value="other_value",
            tags=["new", "tags"],
            overwrite=True,
        )
        assert updated.value == "other_value"
        assert updated.tags == ["new", "tags"]

    test_flow()


async def test_set_async(variable):
    # confirm variable doesn't exist
    variable_doesnt_exist = await Variable.get("my_new_variable")
    assert variable_doesnt_exist is None

    # create new
    created = await Variable.set(
        name="my_new_variable", value="my_value", tags=["123", "456"]
    )
    assert created.value == "my_value"
    assert created.tags == ["123", "456"]

    # try to overwrite
    with pytest.raises(
        ValueError,
        match="Variable 'my_new_variable' already exists. Use `overwrite=True` to update it.",
    ):
        await Variable.set(name="my_new_variable", value="other_value")

    # actually overwrite
    updated = await Variable.set(
        name="my_new_variable",
        value="other_value",
        tags=["new", "tags"],
        overwrite=True,
    )
    assert updated.value == "other_value"
    assert updated.tags == ["new", "tags"]


@pytest.mark.parametrize(
    "value",
    [
        "string-value",
        '"string-value"',
        123,
        12.3,
        True,
        False,
        None,
        {"key": "value"},
        ["value1", "value2"],
        {"key": ["value1", "value2"]},
    ],
)
async def test_json_types(value):
    set_value = await Variable.set("json_variable", value=value)
    assert set_value.value == value


async def test_get(variable):
    # get value
    value = await Variable.get(variable.name)
    assert value == variable.value

    # get value of a variable that doesn't exist
    doesnt_exist = await Variable.get("doesnt_exist")
    assert doesnt_exist is None

    # default is respected if it doesn't exist
    doesnt_exist_default = await Variable.get("doesnt_exist", 42)
    assert doesnt_exist_default == 42


async def test_get_async(variable):
    # get value
    value = await Variable.get(variable.name)
    assert value == variable.value

    # get value of a variable that doesn't exist
    doesnt_exist = await Variable.get("doesnt_exist")
    assert doesnt_exist is None

    # default is respected if it doesn't exist
    doesnt_exist_default = await Variable.get("doesnt_exist", 42)
    assert doesnt_exist_default == 42


async def test_unset(variable):
    # unset a variable
    unset = await Variable.unset(variable.name)
    assert unset is True

    # confirm it's gone
    doesnt_exist = await Variable.get(variable.name)
    assert doesnt_exist is None

    # unset a variable that doesn't exist
    unset_doesnt_exist = await Variable.unset("doesnt_exist")
    assert unset_doesnt_exist is False


async def test_unset_async(variable):
    # unset a variable
    unset = await Variable.unset(variable.name)
    assert unset is True

    # confirm it's gone
    doesnt_exist = await Variable.get(variable.name)
    assert doesnt_exist is None

    # unset a variable that doesn't exist
    unset_doesnt_exist = await Variable.unset("doesnt_exist")
    assert unset_doesnt_exist is False


def test_get_in_sync_flow(variable):
    @flow
    def foo():
        var = Variable.get("my_variable")
        return var

    value = foo()
    assert value == variable.value


async def test_get_in_async_flow(variable):
    @flow
    async def foo():
        var = await Variable.get("my_variable")
        return var

    value = await foo()
    assert value == variable.value


def test_set_in_sync_flow():
    @flow
    def foo():
        return Variable.set("my_variable", value="my-value")

    res = foo()
    assert res
    assert res.value == "my-value"


async def test_set_in_async_flow():
    @flow
    async def foo():
        return await Variable.set("my_variable", value="my-value")

    res = await foo()
    assert res
    assert res.value == "my-value"


async def test_unset_in_sync_flow(variable):
    @flow
    def foo():
        Variable.unset(variable.name)

    foo()
    value = await Variable.get(variable.name)
    assert value is None


async def test_unset_in_async_flow(variable):
    @flow
    async def foo():
        await Variable.unset(variable.name)

    await foo()
    value = await Variable.get(variable.name)
    assert value is None



import asyncio
from asyncio import Queue
from contextlib import asynccontextmanager
from typing import AsyncGenerator, AsyncIterable, Dict, Optional, Set

from prefect.logging import get_logger
from prefect.server.events.filters import EventFilter
from prefect.server.events.schemas.events import ReceivedEvent
from prefect.server.utilities import messaging

logger = get_logger(__name__)

subscribers: Set["Queue[ReceivedEvent]"] = set()
filters: Dict["Queue[ReceivedEvent]", EventFilter] = {}

# The maximum number of message that can be waiting for one subscriber, after which
# new messages will be dropped
SUBSCRIPTION_BACKLOG = 256


@asynccontextmanager
async def subscribed(
    filter: EventFilter,
) -> AsyncGenerator["Queue[ReceivedEvent]", None]:
    queue: "Queue[ReceivedEvent]" = Queue(maxsize=SUBSCRIPTION_BACKLOG)

    subscribers.add(queue)
    filters[queue] = filter

    try:
        yield queue
    finally:
        subscribers.remove(queue)
        del filters[queue]


@asynccontextmanager
async def events(
    filter: EventFilter,
) -> AsyncGenerator[AsyncIterable[Optional[ReceivedEvent]], None]:
    async with subscribed(filter) as queue:

        async def consume() -> AsyncGenerator[Optional[ReceivedEvent], None]:
            while True:
                # Use a brief timeout to allow for cancellation, especially when a
                # client disconnects.  Without a timeout here, a consumer may block
                # forever waiting for a message to be put on the queue, and never notice
                # that their client (like a websocket) has actually disconnected.
                try:
                    event = await asyncio.wait_for(queue.get(), timeout=1)
                except asyncio.TimeoutError:
                    # If the queue is empty, we'll yield to the caller with a None in
                    # order to give it control over what happens next.  This helps with
                    # the outbound websocket, where we want to check if the client is
                    # still connected periodically.
                    yield None
                    continue

                yield event

        yield consume()


@asynccontextmanager
async def distributor() -> AsyncGenerator[messaging.MessageHandler, None]:
    async def message_handler(message: messaging.Message):
        assert message.data

        try:
            assert message.attributes
        except Exception:
            return

        if subscribers:
            event = ReceivedEvent.model_validate_json(message.data)
            for queue in subscribers:
                filter = filters[queue]
                if filter.excludes(event):
                    continue

                try:
                    queue.put_nowait(event)
                except asyncio.QueueFull:
                    continue

    yield message_handler


_distributor_task: Optional[asyncio.Task] = None
_distributor_started: Optional[asyncio.Event] = None


async def start_distributor():
    """Starts the distributor consumer as a global background task"""
    global _distributor_task
    global _distributor_started
    if _distributor_task:
        return

    _distributor_started = asyncio.Event()
    _distributor_task = asyncio.create_task(run_distributor(_distributor_started))
    await _distributor_started.wait()


async def stop_distributor():
    """Stops the distributor consumer global background task"""
    global _distributor_task
    global _distributor_started
    if not _distributor_task:
        return

    task = _distributor_task
    _distributor_task = None
    _distributor_started = None

    task.cancel()
    try:
        await asyncio.shield(task)
    except asyncio.CancelledError:
        pass


class Distributor:
    name: str = "Distributor"

    async def start(self):
        await start_distributor()
        try:
            await _distributor_task
        except asyncio.CancelledError:
            pass

    async def stop(self):
        await stop_distributor()


async def run_distributor(started: asyncio.Event):
    """Runs the distributor consumer forever until it is cancelled"""
    global _distributor_started
    async with messaging.ephemeral_subscription(
        topic="events",
    ) as create_consumer_kwargs:
        started.set()
        async with distributor() as handler:
            consumer = messaging.create_consumer(**create_consumer_kwargs)
            await consumer.run(
                handler=handler,
            )



import asyncio
import uuid

import pytest

from prefect import task
from prefect.task_engine import run_task_async
from prefect.task_runs import TaskRunWaiter


class TestTaskRunWaiter:
    @pytest.fixture(autouse=True)
    def teardown(self):
        yield

        TaskRunWaiter.instance().stop()

    def test_instance_returns_singleton(self):
        assert TaskRunWaiter.instance() is TaskRunWaiter.instance()

    def test_instance_returns_instance_after_stop(self):
        instance = TaskRunWaiter.instance()
        instance.stop()
        assert TaskRunWaiter.instance() is not instance

    @pytest.mark.timeout(20)
    @pytest.mark.usefixtures("use_hosted_api_server")
    async def test_wait_for_task_run(self, prefect_client, emitting_events_pipeline):
        """This test will fail with a timeout error if waiting is not working correctly."""

        @task
        async def test_task():
            await asyncio.sleep(1)

        task_run_id = uuid.uuid4()
        asyncio.create_task(run_task_async(task=test_task, task_run_id=task_run_id))

        await TaskRunWaiter.wait_for_task_run(task_run_id)

        await emitting_events_pipeline.process_events()

        task_run = await prefect_client.read_task_run(task_run_id)
        assert task_run.state.is_completed()

    @pytest.mark.usefixtures("use_hosted_api_server")
    async def test_wait_for_task_run_with_timeout(self):
        @task
        async def test_task():
            await asyncio.sleep(5)

        task_run_id = uuid.uuid4()
        run = asyncio.create_task(
            run_task_async(task=test_task, task_run_id=task_run_id)
        )

        await TaskRunWaiter.wait_for_task_run(task_run_id, timeout=1)

        # TaskRunWaiter stopped waiting before the task finished
        assert not run.done()
        await run

    @pytest.mark.timeout(20)
    @pytest.mark.usefixtures("use_hosted_api_server")
    async def test_non_singleton_mode(self, prefect_client, emitting_events_pipeline):
        waiter = TaskRunWaiter()
        assert waiter is not TaskRunWaiter.instance()

        @task
        async def test_task():
            await asyncio.sleep(1)

        task_run_id = uuid.uuid4()
        asyncio.create_task(run_task_async(task=test_task, task_run_id=task_run_id))

        await waiter.wait_for_task_run(task_run_id)

        await emitting_events_pipeline.process_events()

        task_run = await prefect_client.read_task_run(task_run_id)
        assert task_run.state.is_completed()

        waiter.stop()

    @pytest.mark.timeout(20)
    @pytest.mark.usefixtures("use_hosted_api_server")
    async def test_handles_concurrent_task_runs(
        self, prefect_client, emitting_events_pipeline
    ):
        @task
        async def fast_task():
            await asyncio.sleep(1)

        @task
        async def slow_task():
            await asyncio.sleep(5)

        task_run_id_1 = uuid.uuid4()
        task_run_id_2 = uuid.uuid4()

        asyncio.create_task(run_task_async(task=fast_task, task_run_id=task_run_id_1))
        asyncio.create_task(run_task_async(task=slow_task, task_run_id=task_run_id_2))

        await TaskRunWaiter.wait_for_task_run(task_run_id_1)

        await emitting_events_pipeline.process_events()

        task_run_1 = await prefect_client.read_task_run(task_run_id_1)
        task_run_2 = await prefect_client.read_task_run(task_run_id_2)

        assert task_run_1.state.is_completed()
        assert not task_run_2.state.is_completed()

        await TaskRunWaiter.wait_for_task_run(task_run_id_2)

        await emitting_events_pipeline.process_events()

        task_run_1 = await prefect_client.read_task_run(task_run_id_1)
        task_run_2 = await prefect_client.read_task_run(task_run_id_2)

        assert task_run_1.state.is_completed()
        assert task_run_2.state.is_completed()



"""remove_flowrun_deployment_fk

Revision ID: 7d918a392297
Revises: cfdfec5d7557
Create Date: 2023-03-01 15:46:51.575837

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "7d918a392297"
down_revision = "cfdfec5d7557"
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            ALTER TABLE flow_run DROP CONSTRAINT IF EXISTS fk_flow_run__deployment_id__deployment
            """
        )


def downgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            ALTER TABLE flow_run ADD CONSTRAINT fk_flow_run__deployment_id__deployment FOREIGN KEY(deployment_id) REFERENCES deployment (id) ON DELETE SET NULL;
            """
        )
